import contextlib
import contextvars
from typing import AsyncIterator, Iterator, List, Optional

from dlt.common.configuration import known_sections
from dlt.common.configuration.container import Container
from dlt.common.configuration.exceptions import ContextDefaultCannotBeCreated
from dlt.common.configuration.specs.config_section_context import ConfigSectionContext
from dlt.common.exceptions import (
    PipelineStateNotAvailable,
    SourceSectionNotAvailable,
    ResourceNameNotAvailable,
)
from dlt.common.jsonpath import delete_matches, TAnyJsonPath
from dlt.common.pipeline import TPipelineState, pipeline_state
from dlt.common.typing import DictStrAny, REPattern

from dlt.extract.items import SupportsPipe


def _sources_state(pipeline_state_: Optional[TPipelineState] = None, /) -> DictStrAny:
    if pipeline_state_ is None:
        state, _ = pipeline_state(Container())
    else:
        state = pipeline_state_
    if state is None:
        raise PipelineStateNotAvailable()

    sources_state_: DictStrAny = state.setdefault(known_sections.SOURCES, {})  # type: ignore
    return sources_state_


def source_state(source_state_key: Optional[str] = None) -> DictStrAny:
    """Returns a dictionary with the source-scoped state. Source-scoped state may be shared across the resources of a particular source. Please avoid using source scoped state. Check
    the `resource_state` function for resource-scoped state that is visible within particular resource. Dlt state is preserved across pipeline runs and may be used to implement incremental loads.

    #### Note:
    The source state is a python dictionary-like object that is available within the `@dlt.source` and `@dlt.resource` decorated functions and may be read and written to.
    The data within the state is loaded into destination together with any other extracted data and made automatically available to the source/resource extractor functions when they are run next time.
    When using the state:
    * The source state is scoped to a particular source and will be stored under the source name in the pipeline state
    * It is possible to share state across many sources if they share a schema with the same name
    * Any JSON-serializable values can be written and the read from the state. `dlt` dumps and restores instances of Python bytes, DateTime, Date and Decimal types.
    * The state available in the source decorated function is read only and any changes will be discarded.
    * The state available in the resource decorated function is writable and written values will be available on the next pipeline run
    """
    container = Container()

    # get the source name from the section context
    if not source_state_key:
        with contextlib.suppress(ContextDefaultCannotBeCreated):
            sections_context = container[ConfigSectionContext]
            source_state_key = sections_context.source_state_key

    if not source_state_key:
        raise SourceSectionNotAvailable()

    try:
        state = _sources_state()
    except PipelineStateNotAvailable as e:
        # Reraise with source section
        raise PipelineStateNotAvailable(source_state_key) from e

    return state.setdefault(source_state_key, {})  # type: ignore[no-any-return]


def delete_source_state_keys(
    key: TAnyJsonPath, source_state_: Optional[DictStrAny] = None, /
) -> None:
    """Remove one or more key from the source state.
    The `key` can be any number of keys and/or json paths to be removed.
    """
    state_ = source_state() if source_state_ is None else source_state_
    delete_matches(key, state_)


def resource_state(
    resource_name: str = None, source_state_: Optional[DictStrAny] = None, /
) -> DictStrAny:
    """Returns a dictionary with the resource-scoped state. Resource-scoped state is visible only to resource requesting the access. dlt state is preserved across pipeline runs and may be used to implement incremental loads.

    Note that this function accepts the resource name as optional argument. There are rare cases when `dlt` is not able to resolve resource name due to requesting function
    working in different thread than the main. You'll need to pass the name explicitly when you request resource_state from async functions or functions decorated with @defer.

    Summary:
        The resource state is a python dictionary-like object that is available within the `@dlt.resource` decorated functions and may be read and written to.
        The data within the state is loaded into destination together with any other extracted data and made automatically available to the source/resource extractor functions when they are run next time.
        When using the state:
        * The resource state is scoped to a particular resource requesting it.
        * Any JSON-serializable values can be written and the read from the state. `dlt` dumps and restores instances of Python bytes, DateTime, Date and Decimal types.
        * The state available in the resource decorated function is writable and written values will be available on the next pipeline run

    Example:
        The most typical use case for the state is to implement incremental load.
        >>> @dlt.resource(write_disposition="append")
        >>> def players_games(chess_url, players, start_month=None, end_month=None):
        >>>     checked_archives = dlt.current.resource_state().setdefault("archives", [])
        >>>     archives = players_archives(chess_url, players)
        >>>     for url in archives:
        >>>         if url in checked_archives:
        >>>             print(f"skipping archive {url}")
        >>>             continue
        >>>         else:
        >>>             print(f"getting archive {url}")
        >>>             checked_archives.append(url)
        >>>         # get the filtered archive
        >>>         r = requests.get(url)
        >>>         r.raise_for_status()
        >>>         yield r.json().get("games", [])

    Here we store all the urls with game archives in the state and we skip loading them on next run. The archives are immutable. The state will grow with the coming months (and more players).
    Up to few thousand archives we should be good though.
    Args:
        resource_name (str, optional): forces to use state for a resource with this name. Defaults to None.
        source_state_ (Optional[DictStrAny], optional): Alternative source state. Defaults to None.

    Raises:
        ResourceNameNotAvailable: Raise if used outside of resource context or from a different thread than main

    Returns:
        DictStrAny: State dictionary
    """
    state_ = source_state() if source_state_ is None else source_state_
    # backtrace to find the shallowest resource
    if not resource_name:
        resource_name = get_current_pipe_name()
    return state_.setdefault("resources", {}).setdefault(resource_name, {})  # type: ignore


def reset_resource_state(resource_name: str, source_state_: Optional[DictStrAny] = None, /) -> None:
    """Resets the resource state with name `resource_name` by removing it from `source_state`

    Args:
        resource_name (str): The resource key to reset
        source_state_ (Optional[DictStrAny]): Optional source state dictionary to operate on. Use when working outside source context.
    """
    state_ = source_state() if source_state_ is None else source_state_
    if "resources" in state_ and resource_name in state_["resources"]:
        state_["resources"].pop(resource_name)


def _get_matching_sources(
    pattern: REPattern, pipeline_state: Optional[TPipelineState] = None, /
) -> List[str]:
    """Get all source names in state matching the regex pattern"""
    state_ = _sources_state(pipeline_state)
    return [key for key in state_ if pattern.match(key)]


def _get_matching_resources(
    pattern: REPattern, source_state_: Optional[DictStrAny] = None, /
) -> List[str]:
    """Get all resource names in state matching the regex pattern"""
    state_ = source_state() if source_state_ is None else source_state_
    if "resources" not in state_:
        return []
    return [key for key in state_["resources"] if pattern.match(key)]


_CURRENT_PIPE_CONTEXT: contextvars.ContextVar[Optional[SupportsPipe]] = contextvars.ContextVar(
    "current_pipe", default=None
)
"""Currently executing pipe set for all futures and main thread of execution pipe"""


@contextlib.contextmanager
def pipe_context(pipe: SupportsPipe) -> Iterator[None]:
    """Sync context-manager that sets the variable temporarily."""
    token = _CURRENT_PIPE_CONTEXT.set(pipe)
    try:
        yield
    finally:
        _CURRENT_PIPE_CONTEXT.reset(token)


@contextlib.asynccontextmanager
async def async_pipe_context(pipe: SupportsPipe) -> AsyncIterator[None]:
    """Async variant for `async with` blocks."""
    token = _CURRENT_PIPE_CONTEXT.set(pipe)
    try:
        yield
    finally:
        _CURRENT_PIPE_CONTEXT.reset(token)


def get_current_pipe() -> SupportsPipe:
    """When executed from within dlt.resource decorated function, gets execution pipe"""
    pipe = _CURRENT_PIPE_CONTEXT.get()
    if pipe is None:
        raise ResourceNameNotAvailable()
    return pipe


def get_current_pipe_name() -> str:
    """When executed from within dlt.resource decorated function, gets pipe name associated with current resource

    Pipe name is the same as resource name for all currently known cases.
    """
    pipe = _CURRENT_PIPE_CONTEXT.get()
    if pipe is None:
        raise ResourceNameNotAvailable()
    return pipe.name



import os
from typing import Any, Dict, NamedTuple, Optional, Sequence, Set

from dlt.common.configuration import plugins as _plugins
from dlt.common.configuration.specs.pluggable_run_context import RunContextBase
from dlt.common.runtime.run_context import DOT_DLT, RunContext


class McpFeatures(NamedTuple):
    name: str
    tools: Sequence[Any] = ()
    prompts: Sequence[Any] = ()
    providers: Sequence[Any] = ()


def is_workspace_dir(run_dir: str) -> bool:
    """Checks if `run_dir` contains dlt workspace, this is true if a config file is found"""
    return os.path.isfile(os.path.join(run_dir, DOT_DLT, ".workspace"))


@_plugins.hookimpl(specname="plug_run_context", trylast=False)
def plug_workspace_context_impl(
    run_dir: Optional[str], runtime_kwargs: Optional[Dict[str, Any]]
) -> Optional[RunContextBase]:
    # TODO: if recursive search was requested
    # if runtime_kwargs.get("_look_recursive")
    run_dir = os.path.abspath(run_dir or ".")
    if is_workspace_dir(run_dir):
        # import workspace only when context detected
        from dlt._workspace.profile import DEFAULT_PROFILE, read_profile_pin
        from dlt._workspace.run_context import default_name
        from dlt._workspace._workspace_context import WorkspaceRunContext
        from dlt._workspace._known_env import WORKSPACE__PROFILE

        profile: str = None
        if runtime_kwargs:
            profile = runtime_kwargs.get("profile")
        profile = (
            profile
            or os.environ.get(WORKSPACE__PROFILE)
            or read_profile_pin(RunContext(run_dir))
            or DEFAULT_PROFILE
        )
        return WorkspaceRunContext(default_name(run_dir), run_dir, profile)
    elif runtime_kwargs and runtime_kwargs.get("_required") == "WorkspaceRunContext":
        from dlt._workspace.exceptions import WorkspaceRunContextNotAvailable

        raise WorkspaceRunContextNotAvailable(run_dir)

    return None


@_plugins.hookimpl(specname="plug_mcp")
def plug_mcp_pipeline(features: Set[str]) -> Optional[McpFeatures]:
    """Contribute pipeline-scoped tools: table inspection, SQL queries.

    Activated by the "pipeline" feature. Used by both WorkspaceMCP (features=
    {"workspace", "pipeline"}) and PipelineMCP (features={"pipeline"}).
    """
    if "pipeline" not in features:
        return None

    from dlt._workspace.mcp.tools import data_tools

    _workspace_only = {data_tools.list_pipelines, data_tools.list_profiles}
    tools_list = [t for t in data_tools.__tools__ if t not in _workspace_only]

    return McpFeatures(
        name="pipeline",
        tools=tools_list,
    )


@_plugins.hookimpl(specname="plug_mcp")
def plug_mcp_workspace(features: Set[str]) -> Optional[McpFeatures]:
    """Contribute workspace-level tools: pipeline discovery.

    Activated by the "workspace" feature. Used by WorkspaceMCP only (features=
    {"workspace", "pipeline"}) so users can list available pipelines.
    """
    if "workspace" not in features:
        return None

    from dlt._workspace.mcp.tools import data_tools

    return McpFeatures(
        name="workspace",
        tools=[data_tools.list_pipelines, data_tools.list_profiles, data_tools.get_workspace_info],
    )


@_plugins.hookimpl(specname="plug_mcp")
def plug_mcp_toolkit(features: Set[str]) -> Optional[McpFeatures]:
    """Contribute toolkit discovery tools: list and inspect available toolkits.

    Activated by the "toolkit" feature.
    """
    if "toolkit" not in features:
        return None

    from dlt._workspace.mcp.tools import toolkit_tools

    return McpFeatures(name="toolkit", tools=list(toolkit_tools.__tools__))


@_plugins.hookimpl(specname="plug_mcp")
def plug_mcp_secrets(features: Set[str]) -> Optional[McpFeatures]:
    """Contribute secrets management tools: list, view-redacted, update.

    Activated by the "secrets" feature.
    """
    if "secrets" not in features:
        return None

    from dlt._workspace.mcp.tools import secrets_tools

    return McpFeatures(name="secrets", tools=list(secrets_tools.__tools__))


@_plugins.hookimpl(specname="plug_mcp")
def plug_mcp_context(features: Set[str]) -> Optional[McpFeatures]:
    """Contribute AI context tools: search dlthub sources.

    Activated by the "context" feature.
    """
    if "context" not in features:
        return None

    from dlt._workspace.mcp.tools import context_tools

    return McpFeatures(name="context", tools=list(context_tools.__tools__))


__all__ = [
    "plug_workspace_context_impl",
    "plug_mcp_pipeline",
    "plug_mcp_workspace",
    "plug_mcp_toolkit",
    "plug_mcp_secrets",
    "plug_mcp_context",
]



import os

from dlt.common.configuration.container import Container
from dlt.common.configuration.providers.toml import ConfigTomlProvider
from dlt.common.configuration.specs import RuntimeConfiguration
from dlt.common.configuration.specs import PluggableRunContext
from dlt.common.runtime.anon_tracker import get_anonymous_id

from dlt._workspace.cli import echo as fmt, utils
from dlt._workspace.cli.exceptions import CliCommandException
from dlt._workspace.cli.utils import get_telemetry_status
from dlt._workspace.cli.config_toml_writer import WritableConfigValue, write_values
from dlt._workspace.cli._urls import DLT_TELEMETRY_DOCS_URL


def telemetry_status_command() -> None:
    if get_telemetry_status():
        fmt.echo("Telemetry is %s" % fmt.bold("ENABLED"))
        fmt.echo("Anonymous id %s" % fmt.bold(get_anonymous_id()))
    else:
        fmt.echo("Telemetry is %s" % fmt.bold("DISABLED"))


def change_telemetry_status_command(enabled: bool) -> None:
    from dlt.common.runtime import run_context

    # value to write
    telemetry_value = [
        WritableConfigValue("dlthub_telemetry", bool, enabled, (RuntimeConfiguration.__section__,))
    ]
    # write local config
    # TODO: use designated (main) config provider (for non secret values) ie. taken from run context
    run_ctx = run_context.active()
    config = ConfigTomlProvider(run_ctx.settings_dir)
    if not config.is_empty:
        write_values(config._config_toml, telemetry_value, overwrite_existing=True)
        config.write_toml()

    # write global config
    global_path = run_ctx.global_dir
    os.makedirs(global_path, exist_ok=True)
    config = ConfigTomlProvider(settings_dir=global_path)
    write_values(config._config_toml, telemetry_value, overwrite_existing=True)
    config.write_toml()

    if enabled:
        fmt.echo("Telemetry switched %s" % fmt.bold("ON"))
    else:
        fmt.echo("Telemetry switched %s" % fmt.bold("OFF"))
    # reload config providers
    Container()[PluggableRunContext].reload_providers()


@utils.track_command("telemetry", False)
def telemetry_status_command_wrapper() -> None:
    telemetry_status_command()


@utils.track_command("telemetry_switch", False, "enabled")
def telemetry_change_status_command_wrapper(enabled: bool) -> None:
    try:
        change_telemetry_status_command(enabled)
    except Exception as ex:
        raise CliCommandException(docs_url=DLT_TELEMETRY_DOCS_URL, raiseable_exception=ex)



import functools
import inspect
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

from fastmcp import FastMCP
from fastmcp.prompts import Prompt

from dlt.common import logger
from dlt.common.configuration.plugins import manager
from dlt._workspace.cli.utils import DEFAULT_MCP_FEATURES as _DEFAULT_MCP_FEATURES


# large sentinel set used to discover all registered feature names
_ALL_FEATURE_NAMES = frozenset(
    {
        "workspace",
        "pipeline",
        "toolkit",
        "secrets",
        "context",
        # add future feature names here so they are discoverable
    }
)


def discover_features() -> Tuple[Set[str], Set[str]]:
    """Probe plugin hooks and return (all_available, extra_only) feature names.

    `all_available` is every feature that at least one hook responds to.
    `extra_only` is `all_available - WorkspaceMCP.DEFAULT_FEATURES`.
    """
    available: Set[str] = set()
    for name in _ALL_FEATURE_NAMES:
        results = manager().hook.plug_mcp(features={name})
        if any(r is not None for r in results):
            available.add(name)
    extra = available - WorkspaceMCP.DEFAULT_FEATURES
    return available, extra


def resolve_features(feature_tokens: Optional[List[str]], defaults: Set[str] = None) -> Set[str]:
    """Resolve `+name`, `-name`, `name` tokens against defaults into a feature set."""
    if defaults is None:
        defaults = WorkspaceMCP.DEFAULT_FEATURES
    if not feature_tokens:
        return set(defaults)

    result = set(defaults)
    for item in feature_tokens:
        # tokens may be comma-separated: "--features=-secrets,+context"
        # because argparse treats leading "-" as a flag, use "=" form for removals
        for token in item.split(","):
            token = token.strip()
            if not token:
                continue
            if token.startswith("-"):
                result.discard(token[1:])
            elif token.startswith("+"):
                result.add(token[1:])
            else:
                result.add(token)
    return result


def _curry_pipeline_name(fn: Callable[..., Any], pipeline_name: str) -> Callable[..., Any]:
    """Bind pipeline_name as the first argument, hiding it from the MCP schema."""

    @functools.wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return fn(pipeline_name, *args, **kwargs)

    sig = inspect.signature(fn)
    params = [p for name, p in sig.parameters.items() if name != "pipeline_name"]
    wrapper.__signature__ = sig.replace(parameters=params)  # type: ignore[attr-defined]
    wrapper.__annotations__ = {k: v for k, v in fn.__annotations__.items() if k != "pipeline_name"}
    return wrapper


class DltMCP(FastMCP):
    def __init__(self, name: str, features: Set[str], port: int = 8000, path: str = "/mcp") -> None:
        super().__init__(name=name)
        self._port = port
        self._path = path
        self._features = features
        self._register_features()

    def run(self, transport: str = "streamable-http", **kwargs: Any) -> None:
        run_kwargs: Dict[str, Any] = dict(kwargs)
        if transport != "stdio":
            run_kwargs.setdefault("port", self._port)
            run_kwargs.setdefault("path", self._path)
        super().run(transport=transport, **run_kwargs)

    def _register_features(self) -> None:
        # each plugin hook returns McpFeatures (tools, prompts, providers) or None
        features_list = manager().hook.plug_mcp(features=self._features)
        for mcp_features in features_list:
            if mcp_features is None:
                continue
            self._register_tools(mcp_features.tools)
            # prompts can be raw functions or pre-built Prompt objects
            for prompt_fn in mcp_features.prompts:
                if isinstance(prompt_fn, Prompt):
                    self.add_prompt(prompt_fn)
                else:
                    self.add_prompt(Prompt.from_function(prompt_fn))
            for provider in mcp_features.providers:
                self.add_provider(provider)
        logger.debug("dlt MCP features registered for %s.", self._features)

    def _register_tools(self, tools: List[Any]) -> None:
        for tool in tools:
            self.add_tool(tool)


class WorkspaceMCP(DltMCP):
    """MCP working in Workspace context"""

    DEFAULT_FEATURES: Set[str] = set(_DEFAULT_MCP_FEATURES)

    def __init__(
        self,
        name: str = "dlt",
        port: int = 8000,
        path: str = "/mcp",
        features: Optional[Set[str]] = None,
        extra_features: Optional[Set[str]] = None,
    ) -> None:
        if features is not None:
            resolved = features
        else:
            resolved = self.DEFAULT_FEATURES | (extra_features or set())
        super().__init__(name=name, features=resolved, port=port, path=path)


class PipelineMCP(DltMCP):
    """MCP working in a context of particular pipeline"""

    def __init__(self, pipeline_name: str, port: int = 8000, path: str = "/mcp") -> None:
        self.pipeline_name = pipeline_name
        super().__init__(
            name=f"dlt pipeline: {pipeline_name}",
            features={"pipeline"},
            port=port,
            path=path,
        )

    def _register_tools(self, tools: List[Any]) -> None:
        for tool_fn in tools:
            self.add_tool(_curry_pipeline_name(tool_fn, self.pipeline_name))



import json
from functools import lru_cache, partial
from typing import (
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    cast,
    Any,
)

from dlt.common.data_types.type_helpers import (
    PY_TYPE_TO_SC_TYPE,
    _COERCE_DISPATCH,
    coerce_value,
    py_type_to_sc_type,
)
from dlt.common.data_types.typing import TDataType
from dlt.common.normalizers.exceptions import InvalidJsonNormalizer
from dlt.common.normalizers.typing import TJSONNormalizer
from dlt.common.normalizers.utils import generate_dlt_id
from dlt.common.typing import DictStrAny, TDataItem, StrAny
from dlt.common.schema import Schema
from dlt.common.schema.typing import (
    C_DLT_ID,
    C_DLT_LOAD_ID,
    TColumnName,
    TSimpleRegex,
)
from dlt.common.schema.utils import (
    column_name_validator,
    get_root_table,
    get_first_column_name_with_prop,
)
from dlt.common.utils import update_dict_nested
from dlt.common.normalizers.json import (
    TNormalizedRowIterator,
    wrap_in_dict,
    DataItemNormalizer as DataItemNormalizerBase,
)
from dlt.common.normalizers.json.typing import (
    RelationalNormalizerConfig,
    RelationalNormalizerConfigPropagation,
)
from dlt.common.normalizers.json import helpers as normalize_helpers
from dlt.common.normalizers.json.helpers import (
    get_nested_row_hash,
    get_propagation_mapping,
    get_row_hash,
    requires_root_key,
)
from dlt.common.normalizers.json.expansion import expand_json_column
from dlt.common.validation import validate_dict


class DataItemNormalizer(DataItemNormalizerBase[RelationalNormalizerConfig]):
    # known normalizer props
    C_DLT_ROOT_ID = "_dlt_root_id"
    """unique id of top level parent"""
    C_DLT_PARENT_ID = "_dlt_parent_id"
    """unique id of parent row"""
    C_DLT_LIST_IDX = "_dlt_list_idx"
    """position in the list of rows"""
    C_VALUE = "value"
    """for lists of simple types"""

    # other constants
    EMPTY_KEY_IDENTIFIER = "_empty"  # replace empty keys with this
    RELATIONAL_CONFIG_TYPE: ClassVar[Type[RelationalNormalizerConfig]] = RelationalNormalizerConfig

    normalizer_config: RelationalNormalizerConfig
    propagation_config: RelationalNormalizerConfigPropagation
    max_nesting: int
    _skip_primary_key: Dict[str, bool]

    def __init__(self, schema: Schema) -> None:
        """This item normalizer works with nested dictionaries. It flattens dictionaries and descends into lists.
        It yields row dictionaries at each nesting level."""
        self.schema = schema
        self.naming = schema.naming
        self._load_id: str = None
        self._reset()

    def _reset(self) -> None:
        # normalize known normalizer column identifiers
        self.c_dlt_id: TColumnName = TColumnName(self.naming.normalize_identifier(C_DLT_ID))
        self.c_dlt_load_id: TColumnName = TColumnName(
            self.naming.normalize_identifier(C_DLT_LOAD_ID)
        )
        self.c_dlt_root_id: TColumnName = TColumnName(
            self.naming.normalize_identifier(self.C_DLT_ROOT_ID)
        )
        self.c_dlt_parent_id: TColumnName = TColumnName(
            self.naming.normalize_identifier(self.C_DLT_PARENT_ID)
        )
        self.c_dlt_list_idx: TColumnName = TColumnName(
            self.naming.normalize_identifier(self.C_DLT_LIST_IDX)
        )
        self.c_value: TColumnName = TColumnName(self.naming.normalize_identifier(self.C_VALUE))

        # normalize config

        self.normalizer_config = self.schema._normalizers_config["json"].get("config") or {}  # type: ignore[assignment]
        self.propagation_config = self.normalizer_config.get("propagation", None)
        self.max_nesting = self.normalizer_config.get("max_nesting", 1000)
        self._skip_primary_key = {}
        # create cached versions of helper functions
        self._get_root_row_id_type = lru_cache(maxsize=None)(
            partial(normalize_helpers.get_root_row_id_type, self.schema)
        )
        self._shorten_fragments = lru_cache(maxsize=None)(
            partial(normalize_helpers.shorten_fragments, self.naming)
        )
        self._normalize_table_identifier = lru_cache(maxsize=None)(
            partial(normalize_helpers.normalize_table_identifier, self.schema, self.naming)
        )
        self._normalize_identifier = lru_cache(maxsize=None)(
            partial(normalize_helpers.normalize_identifier, self.schema, self.naming)
        )
        self._get_table_nesting_level = lru_cache(maxsize=None)(
            partial(normalize_helpers.get_table_nesting_level, self.schema)
        )
        self._get_primary_key = lru_cache(maxsize=None)(
            partial(normalize_helpers.get_primary_key, self.schema)
        )
        self._is_nested_type = lru_cache(maxsize=None)(
            partial(normalize_helpers.is_nested_type, self.schema)
        )
        self._should_be_nested = lru_cache(maxsize=None)(
            partial(normalize_helpers.should_be_nested, self.schema)
        )
        self._get_propagation_mapping = (
            lru_cache(maxsize=None)(partial(get_propagation_mapping, self.propagation_config))
            if self.propagation_config
            else None
        )

    @property
    def py_type_to_sc_type_map(self) -> Dict[Type[Any], TDataType]:
        return PY_TYPE_TO_SC_TYPE

    def py_type_to_sc_type(self, t: Type[Any]) -> TDataType:
        return py_type_to_sc_type(t)

    def can_coerce_type(self, to_type: TDataType, from_type: TDataType) -> bool:
        return to_type == from_type or (to_type, from_type) in _COERCE_DISPATCH

    def coerce_type(self, to_type: TDataType, from_type: TDataType, value: Any) -> Any:
        return coerce_value(to_type, from_type, value)

    def _flatten(
        self, table: str, dict_row: DictStrAny, _r_lvl: int
    ) -> Tuple[DictStrAny, Dict[Tuple[str, ...], Sequence[Any]]]:
        out_rec_row: DictStrAny = {}
        out_rec_list: Dict[Tuple[str, ...], Sequence[Any]] = {}

        # cache method lookups as locals
        normalize_identifier = self._normalize_identifier
        normalize_table_identifier = self._normalize_table_identifier
        shorten_fragments = self._shorten_fragments
        is_nested_type = self._is_nested_type
        EMPTY_KEY = self.EMPTY_KEY_IDENTIFIER

        stack: List[Tuple[Dict[str, Any], int, Tuple[str, ...]]] = [(dict_row, _r_lvl, ())]

        while stack:
            current_row, r_lvl, path = stack.pop()
            for k, v in current_row.items():
                norm_k = normalize_identifier(k) if (k and not k.isspace()) else EMPTY_KEY
                nested_name = norm_k if not path else shorten_fragments(*path, norm_k)
                v_type = type(v)
                if v_type is dict:
                    if not is_nested_type(table, nested_name, r_lvl):
                        stack.append((v, r_lvl - 1, path + (norm_k,)))
                        continue
                elif v_type is list:
                    if not is_nested_type(table, nested_name, r_lvl):
                        out_rec_list[path + (normalize_table_identifier(k),)] = v
                        continue

                out_rec_row[nested_name] = v

        return out_rec_row, out_rec_list

    def _add_row_id(
        self,
        table: str,
        dict_row: DictStrAny,
        flattened_row: DictStrAny,
        parent_row_id: str,
        pos: int,
        is_root: bool,
    ) -> str:
        if not is_root:
            row_id = get_nested_row_hash(parent_row_id, table, pos)
            flattened_row[self.c_dlt_parent_id] = parent_row_id
            flattened_row[self.c_dlt_list_idx] = pos
        else:  # root table
            row_id_type = self._get_root_row_id_type(table)
            if row_id_type in ("key_hash", "row_hash"):
                subset = None
                if row_id_type == "key_hash":
                    # primary key based hash must be performed on normalized names
                    subset = self._get_primary_key(table)
                    row_id = get_row_hash(flattened_row, subset=subset)
                else:
                    # base hash on `dict_row` instead of `flattened_row`
                    # so changes in nested tables lead to new row id
                    row_id = get_row_hash(dict_row)
            else:
                row_id = generate_dlt_id()

        flattened_row[self.c_dlt_id] = row_id
        return row_id

    def _get_propagated_values(
        self, table: str, row: DictStrAny, is_root: bool
    ) -> Optional[DictStrAny]:
        mappings = self._get_propagation_mapping(table, is_root)
        if not mappings:
            return None

        extend: DictStrAny = {}
        for prop_from, prop_as in mappings.items():
            if prop_from in row:
                extend[prop_as] = row[prop_from]

        return extend if extend else None

    # generate nested tables only for lists
    def _normalize_list(
        self,
        seq: Sequence[Any],
        extend: DictStrAny,
        ident_path: Tuple[str, ...],
        parent_path: Tuple[str, ...],
        parent_row_id: Optional[str] = None,
        _r_lvl: int = 0,
    ) -> TNormalizedRowIterator:
        # cache method lookups
        shorten_fragments = self._shorten_fragments
        should_be_nested = self._should_be_nested
        c_value = self.c_value

        table = shorten_fragments(*parent_path, *ident_path)
        is_root = not should_be_nested(table)
        for idx, v in enumerate(seq):
            # use type() for exact match - faster than isinstance()
            v_type = type(v)
            if v_type is dict:
                yield from self._normalize_row(
                    v, extend, ident_path, parent_path, parent_row_id, idx, _r_lvl
                )
            elif v_type is list:
                # to normalize lists of lists, we must create a tracking intermediary table
                yield from self._normalize_row(
                    {"list": v},
                    extend,
                    ident_path,
                    parent_path,
                    parent_row_id,
                    idx,
                    _r_lvl - 1,
                )
            else:
                # found non-dict in seq, so wrap it
                wrap_v: DictStrAny = {c_value: v}
                if extend:
                    wrap_v.update(extend)
                self._add_row_id(table, wrap_v, wrap_v, parent_row_id, idx, is_root)
                yield (table, parent_path, ident_path), wrap_v

    def _normalize_row(
        self,
        dict_row: DictStrAny,
        extend: DictStrAny,
        ident_path: Tuple[str, ...],
        parent_path: Tuple[str, ...] = (),
        parent_row_id: Optional[str] = None,
        pos: Optional[int] = None,
        _r_lvl: int = 0,
        is_root: bool = False,
    ) -> TNormalizedRowIterator:
        table = self._shorten_fragments(*parent_path, *ident_path)

        # Store raw JSON values that should be re-injected after flattening.
        originals_to_inject: DictStrAny = {}

        # Apply JSON expansion hints before running the standard flattening logic.
        table_schema = self.schema.tables.get(table)
        if table_schema:
            expanded_row = dict_row.copy()

            for column_name, column_schema in table_schema.get("columns", {}).items():
                if column_name not in dict_row:
                    continue

                flatten_spec = column_schema.get("x-json-flatten")
                keep_original: bool = column_schema.get("x-json-keep-original") == True

                # keep_original with dict but no flatten_spec: store raw JSON on the original column
                if not flatten_spec and keep_original and isinstance(dict_row[column_name], dict):
                    expanded_row[column_name] = json.dumps(dict_row[column_name])
                    continue

                if flatten_spec or keep_original:
                    original_value, expanded_dict = expand_json_column(
                        dict_row[column_name],
                        flatten_spec,
                        keep_original,
                    )

                    # Hand expanded dict back to the standard flattening logic by
                    # setting it as the column value; RelationalNormalizer._flatten
                    # will create the final column names (e.g. column__key).
                    if expanded_dict is not None:
                        expanded_row[column_name] = expanded_dict

                    # Cache raw JSON values to inject back after flattening.
                    if keep_original:
                        originals_to_inject[column_name] = original_value

            dict_row = expanded_row

        # flatten current row and extract all lists to recur into
        flattened_row, lists = self._flatten(table, dict_row, _r_lvl)

        # Inject raw JSON values on original columns after flattening.
        for col, orig_val in originals_to_inject.items():
            flattened_row[col] = orig_val

        # only extend row if there's something to extend
        if extend:
            flattened_row.update(extend)

        # identify load id if loaded data must be processed after loading incrementally
        if is_root := is_root or not self._should_be_nested(table):
            flattened_row[self.c_dlt_load_id] = self._load_id

        # infer record hash or leave existing primary key if present
        row_id = flattened_row.get(self.c_dlt_id)
        if not row_id:
            row_id = self._add_row_id(table, dict_row, flattened_row, parent_row_id, pos, is_root)

        # find fields to propagate to nested tables in config
        if self.propagation_config:
            propagated = self._get_propagated_values(table, flattened_row, is_root)
            if propagated:
                extend = {**extend, **propagated} if extend else propagated

        # yield parent table first
        should_descend = yield (table, parent_path, ident_path), flattened_row
        if should_descend is False:
            return

        # normalize and yield lists - pre-compute parent path once
        if lists:
            new_parent_path = parent_path + ident_path
            for list_path, list_content in lists.items():
                yield from self._normalize_list(
                    list_content,
                    extend,
                    list_path,
                    new_parent_path,
                    row_id,
                    _r_lvl - 1,
                )

    def extend_schema(self, extend_tables: bool = True) -> None:
        """Extends Schema with normalizer-specific hints and settings.

        This method is called by Schema when instance is created or restored from storage.
        """
        config = cast(
            RelationalNormalizerConfig,
            self.schema._normalizers_config["json"].get("config") or {},
        )
        DataItemNormalizer._validate_normalizer_config(self.schema, config)

        # add hints, do not compile.
        self.schema._merge_hints(
            {
                "not_null": [
                    TSimpleRegex(self.c_dlt_id),
                    TSimpleRegex(self.c_dlt_root_id),
                    TSimpleRegex(self.c_dlt_parent_id),
                    TSimpleRegex(self.c_dlt_list_idx),
                    TSimpleRegex(self.c_dlt_load_id),
                ],
                "parent_key": [TSimpleRegex(self.c_dlt_parent_id)],
                "root_key": [TSimpleRegex(self.c_dlt_root_id)],
                "unique": [TSimpleRegex(self.c_dlt_id)],
                "row_key": [TSimpleRegex(self.c_dlt_id)],
            },
            normalize_identifiers=False,  # already normalized
        )

        if extend_tables:
            for table_name in self.schema.tables.keys():
                self.extend_table(table_name)

    def extend_table(self, table_name: str) -> None:
        """If the table has a merge write disposition, add propagation info to normalizer

        Called by Schema when new table is added to schema or table is updated with partial table.
        Table name should be normalized.
        """
        # find root table
        root_table = get_root_table(self.schema.tables, table_name)
        root_table_name = root_table["name"]
        # add root key prop when merge disposition is used or any of nested tables needs row_key
        if requires_root_key(
            self.schema, root_table, self.normalizer_config.get("root_key_propagation")
        ):
            # get row id column from table, assume that we propagate it into c_dlt_root_id always
            c_dlt_id = get_first_column_name_with_prop(
                root_table, "row_key", include_incomplete=True
            )
            self.update_normalizer_config(
                self.schema,
                {
                    "propagation": {
                        "tables": {
                            root_table_name: {
                                TColumnName(c_dlt_id or self.c_dlt_id): TColumnName(
                                    self.c_dlt_root_id
                                )
                            }
                        }
                    }
                },
            )
        # clear all caches that depend on schema.tables since table was added/modified
        self._clear_schema_caches()

    def remove_table(self, table_name: str) -> None:
        """Called by the Schema when table is removed from it."""
        config = self.get_normalizer_config(self.schema)
        if propagation := config.get("propagation"):
            if tables := propagation.get("tables"):
                tables.pop(table_name, None)
        # clear all caches that depend on schema.tables since table was removed
        self._clear_schema_caches()

    def _clear_schema_caches(self) -> None:
        """Clear all lru_caches that depend on schema.tables."""
        self._get_table_nesting_level.cache_clear()
        self._get_primary_key.cache_clear()
        self._is_nested_type.cache_clear()
        self._should_be_nested.cache_clear()
        self._get_root_row_id_type.cache_clear()
        if self._get_propagation_mapping:
            self._get_propagation_mapping.cache_clear()

    def normalize_data_item(
        self, item: TDataItem, load_id: str, table_name: str
    ) -> TNormalizedRowIterator:
        # wrap items that are not dictionaries in dictionary, otherwise they cannot be processed by the JSON normalizer
        if not isinstance(item, dict):
            item = wrap_in_dict(self.c_value, item)
        self._load_id = load_id

        # get table name and nesting level
        root_table_name = self._normalize_table_identifier(table_name)
        max_nesting = self._get_table_nesting_level(root_table_name, self.max_nesting)

        yield from self._normalize_row(
            item,
            {},
            (root_table_name,),
            _r_lvl=max_nesting,  # we count backwards
            is_root=True,
        )

    @classmethod
    def ensure_this_normalizer(cls, norm_config: TJSONNormalizer) -> None:
        # make sure schema has right normalizer
        present_normalizer = norm_config["module"]
        if present_normalizer != cls.__module__:
            raise InvalidJsonNormalizer(cls.__module__, present_normalizer)

    @classmethod
    def update_normalizer_config(cls, schema: Schema, config: RelationalNormalizerConfig) -> None:
        cls._validate_normalizer_config(schema, config)
        existing_config = schema._normalizers_config["json"]
        cls.ensure_this_normalizer(existing_config)
        if "config" in existing_config:
            update_dict_nested(existing_config["config"], config)  # type: ignore
        else:
            existing_config["config"] = config

    @classmethod
    def get_normalizer_config(cls, schema: Schema) -> RelationalNormalizerConfig:
        norm_config = schema._normalizers_config["json"]
        cls.ensure_this_normalizer(norm_config)
        return cast(RelationalNormalizerConfig, norm_config.get("config", {}))

    @classmethod
    def _validate_normalizer_config(
        cls, schema: Schema, config: RelationalNormalizerConfig
    ) -> None:
        """Normalizes all known column identifiers according to the schema and then validates the configuration"""

        def _normalize_prop(
            mapping: Mapping[TColumnName, TColumnName]
        ) -> Dict[TColumnName, TColumnName]:
            return {
                TColumnName(schema.naming.normalize_path(from_col)): TColumnName(
                    schema.naming.normalize_path(to_col)
                )
                for from_col, to_col in mapping.items()
            }

        # normalize the identifiers first
        propagation_config = config.get("propagation")
        if propagation_config:
            if "root" in propagation_config:
                propagation_config["root"] = _normalize_prop(propagation_config["root"])
            if "tables" in propagation_config:
                for table_name in propagation_config["tables"]:
                    propagation_config["tables"][table_name] = _normalize_prop(
                        propagation_config["tables"][table_name]
                    )

        validate_dict(
            cls.RELATIONAL_CONFIG_TYPE,
            config,
            "./normalizers/json/config",
            validator_f=column_name_validator(schema.naming),
        )



from contextlib import contextmanager
from typing import Any, AnyStr, ClassVar, Iterator, List, Optional, Sequence, Generator, Union

import google.cloud.bigquery as bigquery  # noqa: I250
from google.api_core import exceptions as api_core_exceptions
from google.cloud import exceptions as gcp_exceptions
from google.cloud.bigquery import dbapi as bq_dbapi
from google.cloud.bigquery.dbapi import Connection as DbApiConnection, Cursor as BQDbApiCursor
from google.cloud.bigquery.dbapi import exceptions as dbapi_exceptions

from dlt.common import logger
from dlt.common.configuration.specs import (
    GcpServiceAccountCredentialsWithoutDefaults,
    GcpOAuthCredentials,
)
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.typing import StrAny
from dlt.destinations.exceptions import (
    DatabaseTerminalException,
    DatabaseTransientException,
    DatabaseUndefinedRelation,
)
from dlt.destinations.sql_client import (
    DBApiCursorImpl,
    SqlClientBase,
    raise_database_error,
    raise_open_connection_error,
)
from dlt.destinations.typing import DBApi, DBTransaction, DataFrame, ArrowTable
from dlt.common.destination.dataset import DBApiCursor


# terminal reasons as returned in BQ gRPC error response
# https://cloud.google.com/bigquery/docs/error-messages
BQ_TERMINAL_REASONS = [
    "billingTierLimitExceeded",
    "duplicate",
    "invalid",
    "notFound",
    "notImplemented",
    "stopped",
    "tableUnavailable",
]
# invalidQuery is a transient error -> must be fixed by programmer


class BigQueryDBApiCursorImpl(DBApiCursorImpl):
    """Use native BigQuery data frame support if available"""

    native_cursor: BQDbApiCursor  # type: ignore

    def __init__(self, curr: DBApiCursor) -> None:
        super().__init__(curr)

    def iter_df(self, chunk_size: int) -> Generator[DataFrame, None, None]:
        yield from self.native_cursor.query_job.result(page_size=chunk_size).to_dataframe_iterable()

    def iter_arrow(self, chunk_size: int) -> Generator[ArrowTable, None, None]:
        yield from self.native_cursor.query_job.result(page_size=chunk_size).to_arrow_iterable()


class BigQuerySqlClient(SqlClientBase[bigquery.Client], DBTransaction):
    dbapi: ClassVar[DBApi] = bq_dbapi

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: Union[GcpServiceAccountCredentialsWithoutDefaults, GcpOAuthCredentials],
        capabilities: DestinationCapabilitiesContext,
        location: str = "US",
        project_id: Optional[str] = None,
        http_timeout: float = 15.0,
        retry_deadline: float = 60.0,
    ) -> None:
        self._client: bigquery.Client = None
        self.credentials: Union[
            GcpServiceAccountCredentialsWithoutDefaults, GcpOAuthCredentials
        ] = credentials
        self.location = location
        self.project_id = project_id or self.credentials.project_id
        self.http_timeout = http_timeout
        super().__init__(self.project_id, dataset_name, staging_dataset_name, capabilities)

        self._default_retry = bigquery.DEFAULT_RETRY.with_deadline(retry_deadline)
        self._default_query = bigquery.QueryJobConfig(
            default_dataset=self.fully_qualified_dataset_name(quote=False)
        )
        self._session_query: bigquery.QueryJobConfig = None

    @raise_open_connection_error
    def open_connection(self) -> bigquery.Client:
        self._client = bigquery.Client(
            self.project_id,
            credentials=self.credentials.to_native_credentials(),
            location=self.location,
        )

        # patch the client query, so our defaults are used
        query_orig = self._client.query

        def query_patch(
            query: str,
            retry: Any = self._default_retry,
            timeout: Any = self.http_timeout,
            **kwargs: Any,
        ) -> Any:
            return query_orig(query, retry=retry, timeout=timeout, **kwargs)

        self._client.query = query_patch  # type: ignore
        return self._client

    def close_connection(self) -> None:
        if self._session_query:
            self.rollback_transaction()
        if self._client:
            self._client.close()
            self._client = None

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        try:
            if self._session_query:
                raise dbapi_exceptions.ProgrammingError(
                    "Nested transactions not supported on BigQuery"
                )
            job = self._client.query(
                "BEGIN TRANSACTION",
                job_config=bigquery.QueryJobConfig(
                    create_session=True,
                    default_dataset=self.fully_qualified_dataset_name(quote=False),
                ),
            )
            self._session_query = bigquery.QueryJobConfig(
                create_session=False,
                default_dataset=self.fully_qualified_dataset_name(quote=False),
                connection_properties=[
                    bigquery.query.ConnectionProperty(
                        key="session_id", value=job.session_info.session_id
                    )
                ],
            )
            try:
                job.result()
            except Exception:
                # if session creation fails
                self._session_query = None
                raise
            yield self
            self.commit_transaction()
        except Exception:
            self.rollback_transaction()
            raise

    def commit_transaction(self) -> None:
        if not self._session_query:
            # allow committing without transaction
            return
        self.execute_sql("COMMIT TRANSACTION;CALL BQ.ABORT_SESSION()")
        self._session_query = None

    def rollback_transaction(self) -> None:
        if not self._session_query:
            raise dbapi_exceptions.ProgrammingError("Transaction was not started")
        self.execute_sql("ROLLBACK TRANSACTION;CALL BQ.ABORT_SESSION()")
        self._session_query = None

    @property
    def native_connection(self) -> bigquery.Client:
        return self._client

    def has_dataset(self) -> bool:
        try:
            self._client.get_dataset(
                self.fully_qualified_dataset_name(quote=False),
                retry=self._default_retry,
                timeout=self.http_timeout,
            )
            return True
        except gcp_exceptions.NotFound:
            return False

    def create_dataset(self) -> None:
        dataset = bigquery.Dataset(self.fully_qualified_dataset_name(quote=False))
        dataset.location = self.location
        dataset.is_case_insensitive = not self.capabilities.has_case_sensitive_identifiers
        try:
            self._client.create_dataset(
                dataset,
                retry=self._default_retry,
                timeout=self.http_timeout,
            )
        except api_core_exceptions.GoogleAPICallError as gace:
            reason = BigQuerySqlClient._get_reason_from_errors(gace)
            if reason == "notFound":
                # google.api_core.exceptions.NotFound: 404 – table not found
                raise DatabaseUndefinedRelation(gace) from gace
            elif reason in BQ_TERMINAL_REASONS:
                raise DatabaseTerminalException(gace) from gace
            else:
                raise DatabaseTransientException(gace) from gace

    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if not curr.description:
                return None
            try:
                return curr.fetchall()
            except api_core_exceptions.InvalidArgument as ia_ex:
                if "non-table entities cannot be read" in str(ia_ex):
                    return None
                raise

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        conn: DbApiConnection = None
        db_args = args or (kwargs or None)
        try:
            conn = DbApiConnection(client=self._client)
            curr = conn.cursor()
            # if session exists give it a preference
            curr.execute(query, db_args, job_config=self._session_query or self._default_query)
            yield BigQueryDBApiCursorImpl(curr)
        finally:
            if conn:
                # will close all cursors
                conn.close()

    def catalog_name(self, quote: bool = True, casefold: bool = True) -> Optional[str]:
        if casefold:
            project_id = self.capabilities.casefold_identifier(self.project_id)
        else:
            project_id = self.project_id
        if quote:
            project_id = self.capabilities.escape_identifier(project_id)
        return project_id

    @property
    def is_hidden_dataset(self) -> bool:
        """Tells if the dataset associated with sql_client is a hidden dataset.

        Hidden datasets are not present in information schema.
        """
        return self.dataset_name.startswith("_")

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if not cls.is_dbapi_exception(ex):
            return ex
        # google cloud exception in first argument: https://github.com/googleapis/python-bigquery/blob/main/google/cloud/bigquery/dbapi/cursor.py#L205
        cloud_ex = ex.args[0]
        reason = cls._get_reason_from_errors(cloud_ex)
        if reason is None:
            if isinstance(ex, (dbapi_exceptions.DataError, dbapi_exceptions.IntegrityError)):
                return DatabaseTerminalException(ex)
            elif isinstance(ex, dbapi_exceptions.ProgrammingError):
                return DatabaseTransientException(ex)
        if reason == "notFound":
            return DatabaseUndefinedRelation(ex)
        if reason == "invalidQuery" and "was not found" in str(ex) and "Dataset" in str(ex):
            return DatabaseUndefinedRelation(ex)
        if (
            reason == "invalidQuery"
            and "Not found" in str(ex)
            and ("Dataset" in str(ex) or "Table" in str(ex))
        ):
            return DatabaseUndefinedRelation(ex)
        if reason == "accessDenied" and "Dataset" in str(ex) and "not exist" in str(ex):
            return DatabaseUndefinedRelation(ex)
        if reason == "invalidQuery" and (
            "Unrecognized name" in str(ex) or "cannot be null" in str(ex)
        ):
            # unknown column, inserting NULL into required field
            return DatabaseTerminalException(ex)
        if reason in BQ_TERMINAL_REASONS:
            return DatabaseTerminalException(ex)
        # anything else is transient
        return DatabaseTransientException(ex)

    def truncate_tables_if_exist(self, *tables: str) -> None:
        """NOTE: We only truncate tables that exist, for auto-detect schema we don't know which tables exist"""
        statements: List[str] = ["DECLARE table_exists BOOL"]
        for t in tables:
            table_name = self.make_qualified_table_name(t)
            statements.append(
                "SET table_exists = (SELECT COUNT(*) > 0 FROM"
                f" `{self.project_id}.{self.dataset_name}.INFORMATION_SCHEMA.TABLES` WHERE"
                f" table_name = '{t}')"
            )
            truncate_stmt = self._truncate_table_sql(table_name)
            statements.append(f"IF table_exists THEN EXECUTE IMMEDIATE '{truncate_stmt}'; END IF")
        self.execute_many(statements)

    @staticmethod
    def _get_reason_from_errors(gace: api_core_exceptions.GoogleAPICallError) -> Optional[str]:
        errors: List[StrAny] = getattr(gace, "errors", None)
        if errors and isinstance(errors, Sequence):
            return errors[0].get("reason")  # type: ignore
        return None

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, dbapi_exceptions.Error)


class TransactionsNotImplementedError(NotImplementedError):
    def __init__(self) -> None:
        super().__init__(
            "BigQuery does not support transaction management. Instead you may wrap your SQL script"
            " in BEGIN TRANSACTION; ... COMMIT TRANSACTION;"
        )



import os
from types import ModuleType
from typing import Any, Dict, List, Optional

from dlt.common import known_env
from dlt.common.configuration.container import Container
from dlt.common.configuration.providers import EnvironProvider
from dlt.common.configuration.providers.provider import ConfigProvider
from dlt.common.configuration.specs import known_sections
from dlt.common.configuration.specs.pluggable_run_context import (
    ProfilesRunContext,
    PluggableRunContext,
)
from dlt.common.configuration.specs.runtime_configuration import RuntimeConfiguration
from dlt.common.runtime.init import initialize_runtime
from dlt.common.runtime.run_context import (
    DOT_DLT,
    switch_context as _switch_context,
    context_uri,
    global_dir,
)
from dlt.common.typing import copy_sig_ret

from dlt._workspace.configuration import WorkspaceConfiguration, WorkspaceRuntimeConfiguration
from dlt._workspace.exceptions import WorkspaceRunContextNotAvailable
from dlt._workspace.profile import BUILT_IN_PROFILES, DEFAULT_PROFILE, read_profile_pin
from dlt._workspace.providers import ProfileConfigTomlProvider, ProfileSecretsTomlProvider
from dlt._workspace.run_context import (
    DEFAULT_LOCAL_FOLDER,
    DEFAULT_WORKSPACE_WORKING_FOLDER,
    default_working_dir,
    switch_profile as _switch_profile,
)


class WorkspaceRunContext(ProfilesRunContext):
    """A run context with workspace."""

    def __init__(self, name: str, run_dir: str, profile: str):
        self._init_run_dir = run_dir
        self._name = name
        self._profile = profile
        self._data_dir: str = None
        self._local_dir: str = None
        self._global_dir = global_dir()
        self._config: WorkspaceConfiguration = None

    @property
    def name(self) -> str:
        """Defines workspace name which is (normalized) parent folder name"""
        return self._name

    @property
    def global_dir(self) -> str:
        """Directory in which global settings are stored ie ~/.dlt/"""
        return self._global_dir

    @property
    def uri(self) -> str:
        return context_uri(self.name, self.run_dir, self.runtime_kwargs)

    @property
    def run_dir(self) -> str:
        """The default run dir is the current working directory but may be overridden by DLT_PROJECT_DIR env variable."""
        return os.environ.get(known_env.DLT_PROJECT_DIR, self._init_run_dir)

    @property
    def local_dir(self) -> str:
        assert self._local_dir, "local_dir used before workspace configuration got resolved"
        return os.environ.get(known_env.DLT_LOCAL_DIR, self._local_dir)

    @property
    def settings_dir(self) -> str:
        """Returns a path to dlt settings directory. If not overridden it resides in current working directory

        The name of the settings folder is '.dlt'. The path is current working directory '.' but may be overridden by DLT_PROJECT_DIR env variable.
        """
        return os.path.join(self.run_dir, DOT_DLT)

    @property
    def data_dir(self) -> str:
        assert self._data_dir, "data_dir used before workspace configuration got resolved"
        return os.environ.get(known_env.DLT_DATA_DIR, self._data_dir)

    def initial_providers(self) -> List[ConfigProvider]:
        return self._initial_providers(self.profile)

    def initialize_runtime(self, runtime_config: RuntimeConfiguration = None) -> None:
        if runtime_config is not None:
            assert isinstance(runtime_config, WorkspaceRuntimeConfiguration)
            self.config.runtime = runtime_config

        # this also resolves workspace config if necessary
        initialize_runtime(self.name, self.config.runtime)

        # if on runtime, add additional tracker
        if self.runtime_config.run_id:
            from dlt._workspace.helpers.runtime import runtime_artifacts
            from dlt.pipeline import trace

            if runtime_artifacts not in trace.TRACKING_MODULES:
                trace.TRACKING_MODULES.append(runtime_artifacts)

    @property
    def runtime_config(self) -> WorkspaceRuntimeConfiguration:
        return self.config.runtime

    @property
    def config(self) -> WorkspaceConfiguration:
        def _to_run_dir(dir_: Optional[str]) -> Optional[str]:
            if not dir_:
                return None
            return os.path.join(self.run_dir, dir_)

        if self._config is None:
            from dlt.common.configuration.resolve import resolve_configuration

            self._config = resolve_configuration(
                WorkspaceConfiguration(), sections=(known_sections.WORKSPACE,)
            )
            # overwrite name
            if self._config.settings.name:
                self._name = self._config.settings.name

            self._data_dir = (
                _to_run_dir(self._config.settings.working_dir) or self._make_default_working_dir()
            )
            self._local_dir = _to_run_dir(self._config.settings.local_dir) or default_working_dir(
                self.run_dir,
                self.name,
                self.profile,
                DEFAULT_LOCAL_FOLDER,
            )
        return self._config

    @property
    def module(self) -> Optional[ModuleType]:
        try:
            return self.import_run_dir_module(self.run_dir)
        except (ImportError, TypeError):
            return None

    @property
    def runtime_kwargs(self) -> Dict[str, Any]:
        return {"profile": self._profile}

    def get_data_entity(self, entity: str) -> str:
        return os.path.join(self.data_dir, entity)

    def get_run_entity(self, entity: str) -> str:
        """Default run context assumes that entities are defined in root dir"""
        return self.run_dir

    def get_setting(self, setting_path: str) -> str:
        return os.path.join(self.settings_dir, setting_path)

    def plug(self) -> None:
        # create temp and data dirs
        os.makedirs(self.data_dir, exist_ok=True)
        os.makedirs(self.local_dir, exist_ok=True)

    def unplug(self) -> None:
        pass

    def reset_config(self) -> None:
        # Drop resolved configuration to force re-resolve with refreshed providers
        self._config = None
        # no need to initialize the _config anew as it's done in .config property

    # SupportsProfilesOnContext

    @property
    def profile(self) -> str:
        return self._profile

    @property
    def default_profile(self) -> str:
        return DEFAULT_PROFILE

    def available_profiles(self) -> List[str]:
        profiles = list(BUILT_IN_PROFILES.keys())
        if pinned_profile := read_profile_pin(self):
            if pinned_profile not in BUILT_IN_PROFILES:
                profiles.append(pinned_profile)
        return profiles

    def configured_profiles(self) -> List[str]:
        """Returns profiles that have configuration or pipelines.

        A profile is considered configured if:
        - It is the current profile
        - It is the pinned profile
        - It has any toml configuration files (config.toml or secrets.toml with profile prefix)
        - It has pipelines in its working directory

        NOTE: calling this function is relatively expensive as it probes all available profiles
        """
        configured: set[str] = set()

        # current profile is always configured
        configured.add(self.profile)

        # pinned profile is always configured
        if pinned := read_profile_pin(self):
            configured.add(pinned)

        # probe all available profiles
        for profile_name in self.available_profiles():
            if profile_name in configured:
                continue

            # Check if profile has any toml config files
            if self._profile_has_config(profile_name):
                configured.add(profile_name)
                continue

            # Check if profile has any pipelines
            if self._profile_has_pipelines(profile_name):
                configured.add(profile_name)

        return list(configured)

    def _initial_providers(self, profile_name: str) -> List[ConfigProvider]:
        providers = [
            EnvironProvider(),
            ProfileSecretsTomlProvider(self.settings_dir, profile_name, self.global_dir),
            ProfileConfigTomlProvider(self.settings_dir, profile_name, self.global_dir),
        ]
        return providers

    def _make_default_working_dir(self, profile_name: str = None) -> str:
        return default_working_dir(
            self.settings_dir,
            self.name,
            profile_name or self.profile,
            DEFAULT_WORKSPACE_WORKING_FOLDER,
        )

    def _has_default_working_dir(self) -> bool:
        """Checks if current working dir has default layout that includes profiles"""
        return self._data_dir == self._make_default_working_dir()

    def _profile_has_config(self, profile_name: str) -> bool:
        """Check if a profile has any configuration files."""
        # check if any profile-specific files were found
        for provider in self._initial_providers(profile_name):
            for location in provider.present_locations:
                # check if it's a profile-specific file (starts with profile name)
                if os.path.basename(location).startswith(f"{profile_name}."):
                    return True
        return False

    def _profile_has_pipelines(self, profile_name: str) -> bool:
        """Check if a profile has any pipelines in its data directory."""
        # non default layouts can be probed
        if not self._has_default_working_dir():
            return False

        working_dir = self._make_default_working_dir(profile_name)
        pipelines_dir = os.path.join(working_dir, "pipelines")
        try:
            return os.path.isdir(pipelines_dir) and bool(os.listdir(pipelines_dir))
        except OSError:
            return False

    def switch_profile(self, new_profile: str) -> "WorkspaceRunContext":
        return switch_context(self.run_dir, new_profile, required="WorkspaceRunContext")


switch_context = copy_sig_ret(_switch_context, WorkspaceRunContext)(_switch_context)
switch_profile = copy_sig_ret(_switch_profile, WorkspaceRunContext)(_switch_profile)


def active() -> WorkspaceRunContext:
    """Returns currently active Workspace"""
    ctx = Container()[PluggableRunContext].context
    if not isinstance(ctx, WorkspaceRunContext):
        raise WorkspaceRunContextNotAvailable(ctx.run_dir)
    return ctx



from typing import Any, Dict, Type, Union, TYPE_CHECKING, Optional, Sequence

from dlt.common.data_types.typing import TDataType
from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.configuration.specs import AwsCredentials
from dlt.common.data_writers.escape import (
    escape_athena_identifier,
    format_bigquery_datetime_literal,
)
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema.typing import (
    TColumnSchema,
    TColumnType,
    TLoaderMergeStrategy,
    TLoaderReplaceStrategy,
    TTableSchema,
)
from dlt.common.typing import TLoaderFileFormat
from dlt.common.utils import without_none

from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.impl.athena.configuration import (
    DEFAULT_AWS_DATA_CATALOG,
    AthenaClientConfiguration,
)

if TYPE_CHECKING:
    from dlt.destinations.impl.athena.athena import AthenaClient


def athena_merge_strategies_selector(
    supported_merge_strategies: Sequence[TLoaderMergeStrategy],
    /,
    *,
    table_schema: TTableSchema,
) -> Sequence[TLoaderMergeStrategy]:
    if table_schema.get("table_format") == "iceberg":
        return supported_merge_strategies
    else:
        return []


def athena_replace_strategies_selector(
    supported_replace_strategies: Sequence[TLoaderReplaceStrategy],
    /,
    *,
    table_schema: TTableSchema,
) -> Sequence[TLoaderReplaceStrategy]:
    if table_schema.get("table_format") == "iceberg":
        # always from staging table
        return ["insert-from-staging"]
    else:
        # only truncate and insert for regular tables
        return ["truncate-and-insert"]


class AthenaTypeMapper(TypeMapperImpl):
    sct_to_unbound_dbt = {
        "json": "string",
        "text": "string",
        "double": "double",
        "bool": "boolean",
        "date": "date",
        "timestamp": "timestamp",
        "bigint": "bigint",
        "binary": "binary",
        "time": "string",
    }

    sct_to_dbt = {"decimal": "decimal(%i,%i)", "wei": "decimal(%i,%i)"}

    dbt_to_sct = {
        "varchar": "text",
        "double": "double",
        "boolean": "bool",
        "date": "date",
        "timestamp": "timestamp",
        "bigint": "bigint",
        "binary": "binary",
        "varbinary": "binary",
        "decimal": "decimal",
        "tinyint": "bigint",
        "smallint": "bigint",
        "int": "bigint",
    }

    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        # TIME is not supported for parquet on Athena
        if loader_file_format == "parquet" and column["data_type"] == "time":
            raise TerminalValueError(
                "Please convert `datetime.time` objects in your data to `str` or"
                " `datetime.datetime`.",
                "time",
            )

    def to_db_integer_type(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        precision = column.get("precision")
        table_format = table.get("table_format")
        if precision is None:
            return "bigint"
        if precision <= 8:
            return "int" if table_format == "iceberg" else "tinyint"
        elif precision <= 16:
            return "int" if table_format == "iceberg" else "smallint"
        elif precision <= 32:
            return "int"
        elif precision <= 64:
            return "bigint"
        raise TerminalValueError(
            f"bigint with `{precision=:}` can't be mapped to athena integer type"
        )

    def from_destination_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        for key, val in self.dbt_to_sct.items():
            if db_type.startswith(key):
                return without_none(dict(data_type=val, precision=precision, scale=scale))  # type: ignore[return-value]
        return dict(data_type=None)


class athena(Destination[AthenaClientConfiguration, "AthenaClient"]):
    spec = AthenaClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        # athena only supports loading from staged files on s3 for now
        caps.preferred_loader_file_format = None
        caps.supported_loader_file_formats = []
        caps.supported_table_formats = ["iceberg", "hive"]
        caps.preferred_staging_file_format = "parquet"
        caps.supported_staging_file_formats = ["parquet", "model"]
        caps.type_mapper = AthenaTypeMapper

        # athena is storing all identifiers in lower case and is case insensitive
        # it also uses lower case in all the queries
        # https://docs.aws.amazon.com/athena/latest/ug/tables-databases-columns-names.html
        caps.escape_identifier = escape_athena_identifier
        caps.casefold_identifier = str.lower
        caps.has_case_sensitive_identifiers = False
        caps.format_datetime_literal = format_bigquery_datetime_literal
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)
        caps.max_identifier_length = 255
        caps.max_column_identifier_length = 255
        caps.max_query_length = 16 * 1024 * 1024
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 262144
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_ddl_transactions = False
        caps.supports_transactions = False
        caps.alter_add_multi_column = True
        caps.schema_supports_numeric_precision = False
        caps.timestamp_precision = 3
        caps.max_timestamp_precision = 3
        caps.supports_truncate_command = False
        caps.supported_merge_strategies = ["delete-insert", "upsert", "scd2"]
        caps.supported_replace_strategies = ["truncate-and-insert", "insert-from-staging"]
        caps.merge_strategies_selector = athena_merge_strategies_selector
        caps.replace_strategies_selector = athena_replace_strategies_selector
        caps.enforces_nulls_on_alter = False
        caps.sqlglot_dialect = "athena"
        caps.supports_tz_aware_datetime = False

        return caps

    @property
    def client_class(self) -> Type["AthenaClient"]:
        from dlt.destinations.impl.athena.athena import AthenaClient

        return AthenaClient

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: AthenaClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        if config.force_iceberg:
            caps.preferred_table_format = "iceberg"

        if config._is_s3_tables_catalog():
            caps.naming_convention = "s3_tables"

        return super().adjust_capabilities(caps, config, naming)

    def __init__(
        self,
        query_result_bucket: Optional[str] = None,
        credentials: Union[AwsCredentials, Dict[str, Any], Any] = None,
        athena_work_group: str = None,
        aws_data_catalog: str = DEFAULT_AWS_DATA_CATALOG,
        staging_aws_data_catalog: str = None,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the Athena destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            query_result_bucket (Optional[str], optional): S3 bucket to store query results in. May be omitted when using Athena managed query results.
            credentials (Union[AwsCredentials, Dict[str, Any], Any], optional): AWS credentials to connect to the Athena database. Can be an instance of `AwsCredentials` or
                a dict with AWS credentials
            athena_work_group (str, optional): Athena work group to use
            aws_data_catalog (str, optional): Catalog used to register production (non-staging) tables in
            staging_aws_data_catalog (str, optional): Catalog used to register staging tables in.
                Should not be an S3 Tables Catalog. If not set, defaults to configured value for
                `aws_data_catalog` if it's not an S3 Tables Catalog, and to `awsdatacatalog` if it is
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any): Additional arguments passed to the destination config
        """
        super().__init__(
            query_result_bucket=query_result_bucket,
            credentials=credentials,
            athena_work_group=athena_work_group,
            aws_data_catalog=aws_data_catalog,
            staging_aws_data_catalog=staging_aws_data_catalog,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )


athena.register()



import os
import inspect
from types import ModuleType
from functools import update_wrapper, wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)
from typing_extensions import TypeVar, Self

from dlt.common import logger
from dlt.common.configuration import with_config, get_fun_spec, known_sections
from dlt.common.configuration.container import Container
from dlt.common.configuration.exceptions import ContextDefaultCannotBeCreated
from dlt.common.configuration.inject import set_fun_spec
from dlt.common.configuration.resolve import inject_section
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.configuration.specs.config_section_context import ConfigSectionContext
from dlt.common.exceptions import ArgumentsOverloadException
from dlt.common.pipeline import PipelineContext
from dlt.common.reflection.spec import spec_from_signature
from dlt.common.reflection.inspect import iscoroutinefunction
from dlt.common.schema.utils import DEFAULT_WRITE_DISPOSITION
from dlt.common.schema.schema import Schema
from dlt.common.schema.typing import (
    TFileFormat,
    TWriteDisposition,
    TWriteDispositionConfig,
    TAnySchemaColumns,
    TSchemaContract,
    TTableFormat,
    TTableReferenceParam,
)
from dlt.common.storages.exceptions import SchemaNotFoundError
from dlt.common.storages.schema_storage import SchemaStorage
from dlt.common.typing import (
    AnyFun,
    ParamSpec,
    Generic,
    Concatenate,
    TDataItem,
    TDataItems,
    TColumnNames,
    TTableNames,
)
from dlt.common.utils import (
    get_callable_name,
    get_module_name,
    is_inner_callable,
    get_full_callable_name,
)

from dlt.extract.hints import TResourceNestedHints, make_hints
from dlt.extract.state import get_current_pipe
from dlt.extract.utils import dynstr
from dlt.extract.exceptions import (
    CurrentSourceNotAvailable,
    InvalidTransformerDataTypeGeneratorFunctionRequired,
    ResourceFunctionExpected,
    SourceDataIsNone,
    SourceIsAClassTypeError,
    ExplicitSourceNameInvalid,
    SourceNotAFunction,
    CurrentSourceSchemaNotAvailable,
)
from dlt.extract.items import TTableHintTemplate
from dlt.extract.source import (
    DltSource,
    SourceSchemaInjectableContext,
    SourceInjectableContext,
    _DltSingleSource,
)
from dlt.extract.reference import SourceReference, SourceFactory, TDltSourceImpl, TSourceFunParams
from dlt.extract.resource import DltResource, TUnboundDltResource, TDltResourceImpl
from dlt.extract.incremental import TIncrementalConfig


class DltSourceFactoryWrapper(SourceFactory[TSourceFunParams, TDltSourceImpl]):
    def __init__(
        self,
    ) -> None:
        """Creates a wrapper that is returned by @source decorator. It preserves the decorated function when called and
        allows to change the decorator arguments at runtime. Changing the `name` and `section` creates a clone of the source
        with different name and taking the configuration from a different keys.

        This wrapper registers the source under `section`.`name` type in SourceReference registry, using the original
        `section` (which corresponds to module name) and `name` (which corresponds to source function name).
        """
        self._f: AnyFun = None
        self.ref: SourceReference = None
        self._deco_f: Callable[..., TDltSourceImpl] = None

        self.name: str = None
        self.section: str = None
        self.max_table_nesting: int = None
        self.root_key: bool = None
        self.schema: Schema = None
        self.schema_contract: TSchemaContract = None
        self.spec: Type[BaseConfiguration] = None
        self.parallelized: bool = None
        self._impl_cls: Type[TDltSourceImpl] = DltSource  # type: ignore[assignment]
        self._postprocessors: List[
            Callable[[TDltSourceImpl], Union[TDltSourceImpl, Awaitable[TDltSourceImpl]]]
        ] = []

    def add_postprocessor(
        self, func: Callable[[TDltSourceImpl], Union[TDltSourceImpl, Awaitable[TDltSourceImpl]]]
    ) -> None:
        """Adds a callback that receives and returns a DltSource after it is created."""
        self._postprocessors.append(func)

    def _apply_postprocessors(self, source: TDltSourceImpl) -> TDltSourceImpl:
        for func in self._postprocessors:
            source = func(source)  # type: ignore[assignment]
        return source

    def clone(
        self,
        *,
        name: str = None,
        section: str = None,
        max_table_nesting: int = None,
        root_key: bool = None,
        schema: Schema = None,
        schema_contract: TSchemaContract = None,
        spec: Type[BaseConfiguration] = None,
        parallelized: bool = None,
        _impl_cls: Type[TDltSourceImpl] = None,
    ) -> Self:
        """Overrides default arguments that will be used to create DltSource instance when this wrapper is called. This method
        clones this wrapper.
        """
        # if source function not set, apply args in place
        ovr = self.__class__() if self._f else self

        if name is not None:
            ovr.name = name
        else:
            ovr.name = self.name
        if section is not None:
            ovr.section = section
        else:
            ovr.section = self.section
        if max_table_nesting is not None:
            ovr.max_table_nesting = max_table_nesting
        else:
            ovr.max_table_nesting = self.max_table_nesting
        if root_key is not None:
            ovr.root_key = root_key
        else:
            ovr.root_key = self.root_key
        ovr.schema = schema or self.schema
        if schema_contract is not None:
            ovr.schema_contract = schema_contract
        else:
            ovr.schema_contract = self.schema_contract
        ovr.spec = spec or self.spec
        if parallelized is not None:
            ovr.parallelized = parallelized
        else:
            ovr.parallelized = self.parallelized
        ovr._impl_cls = _impl_cls or self._impl_cls

        # also remember original source function
        ovr._f = self._f
        ovr.ref = self.ref
        ovr._postprocessors = list(self._postprocessors)
        ovr._update_wrapper()
        # try to bind _f
        ovr.wrap()
        return ovr

    with_args = clone

    def __call__(self, *args: Any, **kwargs: Any) -> TDltSourceImpl:
        assert self._deco_f, f"Attempt to call source function on {self.name} before bind"
        # if source impl is a single resource source
        if issubclass(self._impl_cls, _DltSingleSource):
            # call special source function that will create renamed resource
            source = self._deco_f(self.name, self.section, args, kwargs)
            assert isinstance(source, _DltSingleSource)
            # set source section to empty to not interfere with resource sections, same thing we do in extract
            source.section = source.single_resource.section
            # apply selected settings directly to resource
            resource = source.single_resource
            if self.max_table_nesting is not None:
                resource.max_table_nesting = self.max_table_nesting
            if self.schema_contract is not None:
                resource.apply_hints(schema_contract=self.schema_contract)
        else:
            source = self._deco_f(*args, **kwargs)
        return self._apply_postprocessors(source)

    def bind(self, f: AnyFun) -> Self:
        """Binds wrapper to the original source function and registers the source reference. This method is called only once by the decorator"""
        self._f = f
        self._update_wrapper()
        self.ref = self.wrap()
        if not is_inner_callable(f):
            SourceReference.register(self.ref)
        return self

    def wrap(self) -> SourceReference:
        """Wrap the original source function using _deco."""
        if not self._f:
            return None
        if hasattr(self._f, "__qualname__"):
            self.__qualname__ = self._f.__qualname__
        return self._wrap(self._f)

    def _update_wrapper(self) -> None:
        """wrap self so we preserve signature, module etc. from f"""
        if not callable(self._f):
            return None
        update_wrapper(self, self._f)
        # also change the signature return annotation to dlt source class
        sig = inspect.signature(self._f).replace(return_annotation=self._impl_cls)
        self.__signature__ = sig

    def _wrap(self, f: AnyFun) -> SourceReference:
        """Wraps source function `f` in configuration injector."""
        if not callable(f) or isinstance(f, DltResource):
            raise SourceNotAFunction(self.name or "<no name>", f, type(f))

        if inspect.isclass(f):
            raise SourceIsAClassTypeError(self.name or "<no name>", f)

        # source name is passed directly or taken from decorated function name
        effective_name = self.name or get_callable_name(f)

        if self.schema and self.name and self.name != self.schema.name:
            raise ExplicitSourceNameInvalid(self.name, self.schema.name)

        # wrap source extraction function in configuration with section
        func_module = inspect.getmodule(f)
        source_section = self.section or _get_source_section_name(func_module)
        # use effective_name which is explicit source name or callable name to represent third element in source config path
        source_sections = (known_sections.SOURCES, source_section, effective_name)
        conf_f = with_config(f, spec=self.spec, sections=source_sections)

        def _eval_rv(_rv: Any, schema_copy: Schema) -> TDltSourceImpl:
            """Evaluates return value from the source function or coroutine"""
            if _rv is None:
                raise SourceDataIsNone(schema_copy.name)
            # if generator, consume it immediately
            if inspect.isgenerator(_rv):
                _rv = list(_rv)

            # convert to source
            s = self._impl_cls.from_data(schema_copy, source_section, _rv)
            # apply hints
            if self.max_table_nesting is not None:
                s.max_table_nesting = self.max_table_nesting
            s.schema_contract = self.schema_contract
            # enable root propagation
            s.root_key = self.root_key
            # parallelize resources
            if self.parallelized:
                s.parallelize()
            return s

        def _make_schema() -> Schema:
            if not self.schema:
                # load the schema from file with name_schema.yaml/json from the same directory, the callable resides OR create new default schema
                return _maybe_load_schema_for_callable(f, effective_name) or Schema(effective_name)
            else:
                # clone the schema passed to decorator, update normalizers, remove processing hints
                # NOTE: source may be called several times in many different settings
                return self.schema.clone(update_normalizers=True, remove_processing_hints=True)

        @wraps(conf_f)
        def _wrap(*args: Any, **kwargs: Any) -> TDltSourceImpl:
            """Wrap a regular function, injection context must be a part of the wrap"""
            schema_copy = _make_schema()
            with Container().injectable_context(SourceSchemaInjectableContext(schema_copy)):
                # configurations will be accessed in this section in the source
                proxy = Container()[PipelineContext]
                pipeline_name = None if not proxy.is_active() else proxy.pipeline().pipeline_name
                with inject_section(
                    ConfigSectionContext(
                        pipeline_name=pipeline_name,
                        sections=source_sections,
                        source_state_key=schema_copy.name,
                    )
                ):
                    rv = conf_f(*args, **kwargs)
                    return _eval_rv(rv, schema_copy)

        @wraps(conf_f)
        async def _wrap_coro(*args: Any, **kwargs: Any) -> TDltSourceImpl:
            """In case of co-routine we must wrap the whole injection context in awaitable,
            there's no easy way to avoid some code duplication
            """
            schema_copy = _make_schema()
            with Container().injectable_context(SourceSchemaInjectableContext(schema_copy)):
                # configurations will be accessed in this section in the source
                proxy = Container()[PipelineContext]
                pipeline_name = None if not proxy.is_active() else proxy.pipeline().pipeline_name
                with inject_section(
                    ConfigSectionContext(
                        pipeline_name=pipeline_name,
                        sections=source_sections,
                        source_state_key=schema_copy.name,
                    )
                ):
                    rv = await conf_f(*args, **kwargs)
                    return _eval_rv(rv, schema_copy)

        # get spec for wrapped function
        SPEC = get_fun_spec(conf_f)
        # get correct wrapper
        self._deco_f = _wrap_coro if iscoroutinefunction(f) else _wrap  # type: ignore[assignment]
        return SourceReference(get_full_callable_name(f), SPEC, self, source_section, effective_name)  # type: ignore[arg-type]


TResourceFunParams = ParamSpec("TResourceFunParams")


@overload
def source(
    func: Callable[TSourceFunParams, Any],
    /,
    name: str = None,
    section: str = None,
    max_table_nesting: int = None,
    root_key: bool = False,
    schema: Schema = None,
    schema_contract: TSchemaContract = None,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    _impl_cls: Type[TDltSourceImpl] = DltSource,  # type: ignore[assignment]
) -> SourceFactory[TSourceFunParams, TDltSourceImpl]: ...


@overload
def source(
    func: None = ...,
    /,
    name: str = None,
    section: str = None,
    max_table_nesting: int = None,
    root_key: bool = False,
    schema: Schema = None,
    schema_contract: TSchemaContract = None,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    _impl_cls: Type[TDltSourceImpl] = DltSource,  # type: ignore[assignment]
) -> Callable[
    [Callable[TSourceFunParams, Any]], SourceFactory[TSourceFunParams, TDltSourceImpl]
]: ...


def source(
    func: Optional[AnyFun] = None,
    /,
    name: str = None,
    section: str = None,
    max_table_nesting: int = None,
    root_key: bool = None,
    schema: Schema = None,
    schema_contract: TSchemaContract = None,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    _impl_cls: Type[TDltSourceImpl] = DltSource,  # type: ignore[assignment]
) -> Any:
    """A decorator that transforms a function returning one or more `dlt resources` into a `dlt source` in order to load it with `dlt`.

    Note:
    A `dlt source` is a logical grouping of resources that are often extracted and loaded together. A source is associated with a schema, which describes the structure of the loaded data and provides instructions how to load it.
    Such schema contains table schemas that describe the structure of the data coming from the resources.

    Please refer to https://dlthub.com/docs/general-usage/source for a complete documentation.

    #### Credentials:
    Another important function of the source decorator is to provide credentials and other configuration to the code that extracts data. The decorator may automatically bind the source function arguments to the secret and config values.
    >>> @dlt.source
    >>> def chess(username, chess_url: str = dlt.config.value, api_secret = dlt.secrets.value, title: str = "GM"):
    >>>     return user_profile(username, chess_url, api_secret), user_games(username, chess_url, api_secret, with_titles=title)
    >>>
    >>> list(chess("magnuscarlsen"))

    Here `username` is a required, explicit python argument, `chess_url` is a required argument, that if not explicitly passed will be taken from configuration ie. `config.toml`, `api_secret` is a required argument, that if not explicitly passed will be taken from dlt secrets ie. `secrets.toml`.
    See https://dlthub.com/docs/general-usage/credentials/ for details.

    Args:
        func (Optional[AnyFun]): A function that returns a dlt resource or a list of those or a list of any data items that can be loaded by `dlt`.

        name (str, optional): A name of the source which is also the name of the associated schema. If not present, the function name will be used.

        section (str, optional): Configuration section that comes right after 'sources` in default layout. If not present, the current python module name will be used.
            Default layout is `sources.<section>.<name>.<key_name>`.

        max_table_nesting (int, optional): A schema hint that sets the maximum depth of nested table above which the remaining nodes are loaded as structs or JSON.

        root_key (bool): Enables merging on all resources by propagating row key from root to all nested tables. This option is most useful if you plan to change write disposition of a resource to disable/enable merge. Defaults to False.

        schema (Schema, optional): An explicit `Schema` instance to be associated with the source. If not present, `dlt` creates a new `Schema` object with provided `name`. If such `Schema` already exists in the same folder as the module containing the decorated function, such schema will be loaded from file.

        schema_contract (TSchemaContract, optional): Schema contract settings that will be applied to this resource.

        spec (Type[BaseConfiguration], optional): A specification of configuration and secret values required by the source.

        parallelized (bool, optional): If `True`, resource generators will be extracted in parallel with other resources.
            Transformers that return items are also parallelized. Non-eligible resources are ignored. Defaults to `False` which preserves resource settings.

        _impl_cls (Type[TDltSourceImpl], optional): A custom implementation of DltSource, may be also used to providing just a typing stub

    Returns:
        Any: Wrapped decorated source function, see SourceFactory reference for additional wrapper capabilities
    """
    if name and schema:
        raise ArgumentsOverloadException(
            "`name` has no effect when `schema` argument is present", source.__name__
        )

    source_wrapper = (
        DltSourceFactoryWrapper[Any, TDltSourceImpl]()
        .clone(
            name=name,
            section=section,
            max_table_nesting=max_table_nesting,
            root_key=root_key,
            schema=schema,
            schema_contract=schema_contract,
            spec=spec,
            parallelized=parallelized,
            _impl_cls=_impl_cls,
        )
        .bind
    )

    if func is None:
        # we're called with parens.
        return source_wrapper
    # we're called as @source without parens.
    return source_wrapper(func)


class ResourceFactory(DltResource, Generic[TResourceFunParams, TDltResourceImpl]):
    # this class is used only for typing, do not instantiate, do not add docstring
    def __call__(  # type: ignore[override]
        self, *args: TResourceFunParams.args, **kwargs: TResourceFunParams.kwargs
    ) -> TDltResourceImpl:
        pass


@overload
def resource(
    data: Callable[TResourceFunParams, Any],
    /,
    name: TTableHintTemplate[str] = None,
    table_name: TTableHintTemplate[str] = None,
    max_table_nesting: int = None,
    write_disposition: TTableHintTemplate[TWriteDispositionConfig] = None,
    columns: TTableHintTemplate[TAnySchemaColumns] = None,
    primary_key: TTableHintTemplate[TColumnNames] = None,
    merge_key: TTableHintTemplate[TColumnNames] = None,
    schema_contract: TTableHintTemplate[TSchemaContract] = None,
    table_format: TTableHintTemplate[TTableFormat] = None,
    file_format: TTableHintTemplate[TFileFormat] = None,
    references: TTableHintTemplate[TTableReferenceParam] = None,
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    incremental: Optional[TIncrementalConfig] = None,
    _impl_cls: Type[TDltResourceImpl] = DltResource,  # type: ignore[assignment]
    section: Optional[TTableHintTemplate[str]] = None,
    _base_spec: Type[BaseConfiguration] = BaseConfiguration,
    standalone: bool = None,
) -> ResourceFactory[TResourceFunParams, TDltResourceImpl]: ...


@overload
def resource(
    data: None = ...,
    /,
    name: TTableHintTemplate[str] = None,
    table_name: TTableHintTemplate[str] = None,
    max_table_nesting: int = None,
    write_disposition: TTableHintTemplate[TWriteDispositionConfig] = None,
    columns: TTableHintTemplate[TAnySchemaColumns] = None,
    primary_key: TTableHintTemplate[TColumnNames] = None,
    merge_key: TTableHintTemplate[TColumnNames] = None,
    schema_contract: TTableHintTemplate[TSchemaContract] = None,
    table_format: TTableHintTemplate[TTableFormat] = None,
    file_format: TTableHintTemplate[TFileFormat] = None,
    references: TTableHintTemplate[TTableReferenceParam] = None,
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    incremental: Optional[TIncrementalConfig] = None,
    _impl_cls: Type[TDltResourceImpl] = DltResource,  # type: ignore[assignment]
    section: Optional[TTableHintTemplate[str]] = None,
    _base_spec: Type[BaseConfiguration] = BaseConfiguration,
    standalone: bool = None,
) -> Callable[
    [Callable[TResourceFunParams, Any]], ResourceFactory[TResourceFunParams, TDltResourceImpl]
]: ...


@overload
def resource(
    data: Union[List[Any], Iterator[Any]],
    /,
    name: str = None,
    table_name: TTableHintTemplate[str] = None,
    max_table_nesting: int = None,
    write_disposition: TTableHintTemplate[TWriteDispositionConfig] = None,
    columns: TTableHintTemplate[TAnySchemaColumns] = None,
    primary_key: TTableHintTemplate[TColumnNames] = None,
    merge_key: TTableHintTemplate[TColumnNames] = None,
    schema_contract: TTableHintTemplate[TSchemaContract] = None,
    table_format: TTableHintTemplate[TTableFormat] = None,
    file_format: TTableHintTemplate[TFileFormat] = None,
    references: TTableHintTemplate[TTableReferenceParam] = None,
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    incremental: Optional[TIncrementalConfig] = None,
    _impl_cls: Type[TDltResourceImpl] = DltResource,  # type: ignore[assignment]
    section: Optional[str] = None,
    _base_spec: Type[BaseConfiguration] = BaseConfiguration,
    standalone: bool = None,
) -> TDltResourceImpl: ...


def resource(
    data: Optional[Any] = None,
    /,
    name: TTableHintTemplate[str] = None,
    table_name: TTableHintTemplate[str] = None,
    max_table_nesting: int = None,
    write_disposition: TTableHintTemplate[TWriteDispositionConfig] = None,
    columns: TTableHintTemplate[TAnySchemaColumns] = None,
    primary_key: TTableHintTemplate[TColumnNames] = None,
    merge_key: TTableHintTemplate[TColumnNames] = None,
    schema_contract: TTableHintTemplate[TSchemaContract] = None,
    table_format: TTableHintTemplate[TTableFormat] = None,
    file_format: TTableHintTemplate[TFileFormat] = None,
    references: TTableHintTemplate[TTableReferenceParam] = None,
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    incremental: Optional[TIncrementalConfig] = None,
    _impl_cls: Type[TDltResourceImpl] = DltResource,  # type: ignore[assignment]
    section: Optional[TTableHintTemplate[str]] = None,
    _base_spec: Type[BaseConfiguration] = BaseConfiguration,
    standalone: bool = None,
    data_from: TUnboundDltResource = None,
) -> Any:
    """When used as a decorator, transforms any generator (yielding) function into a `dlt resource`. When used as a function, it transforms data in `data` argument into a `dlt resource`.

    #### Note:
    A `resource`is a location within a `source` that holds the data with specific structure (schema) or coming from specific origin. A resource may be a rest API endpoint, table in the database or a tab in Google Sheets.
    A `dlt resource` is python representation of a `resource` that combines both data and metadata (table schema) that describes the structure and instructs the loading of the data.
    A `dlt resource` is also an `Iterable` and can used like any other iterable object ie. list or tuple.

    Please refer to https://dlthub.com/docs/general-usage/resource for a complete documentation.

    #### Credentials:
    If used as a decorator (`data` argument is a `Generator`), it may automatically bind the source function arguments to the secret and config values.
    >>> @dlt.resource
    >>> def user_games(username, chess_url: str = dlt.config.value, api_secret = dlt.secrets.value):
    >>>     return requests.get("%s/games/%s" % (chess_url, username), headers={"Authorization": f"Bearer {api_secret}"})
    >>>
    >>> list(user_games("magnuscarlsen"))

    Here `username` is a required, explicit python argument, `chess_url` is a required argument, that if not explicitly passed will be taken from configuration ie. `config.toml`, `api_secret` is a required argument, that if not explicitly passed will be taken from dlt secrets ie. `secrets.toml`.
    See https://dlthub.com/docs/general-usage/credentials/ for details.
    Note that if decorated function is an inner function, passing of the credentials will be disabled.

    Args:
        data (Optional[Any], optional): a function to be decorated or a data compatible with `dlt` `run`.

        name (TTableHintTemplate[str], optional): A name of the resource that by default also becomes the name of the table to which the data is loaded.
            If not present, the name of the decorated function will be used.

        table_name (TTableHintTemplate[str], optional): An table name, if different from `name`.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        max_table_nesting (int, optional): A schema hint that sets the maximum depth of nested table above which the remaining nodes are loaded as structs or JSON.

        write_disposition (TTableHintTemplate[TWriteDispositionConfig], optional): Controls how to write data to a table. Accepts a shorthand string literal or configuration dictionary.
            Allowed shorthand string literals: `append` will always add new data at the end of the table. `replace` will replace existing data with new data. `skip` will prevent data from loading. "merge" will deduplicate and merge data based on "primary_key" and "merge_key" hints. Defaults to "append".
            Write behaviour can be further customized through a configuration dictionary. For example, to obtain an SCD2 table provide `write_disposition={"disposition": "merge", "strategy": "scd2"}`.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        columns (TTableHintTemplate[TAnySchemaColumns], optional): A list, dict or pydantic model of column schemas.
            Typed dictionary describing column names, data types, write disposition and performance hints that gives you full control over the created table schema.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.
            When the argument is a pydantic model, the model will be used to validate the data yielded by the resource as well.

        primary_key (TTableHintTemplate[TColumnNames], optional): A column name or a list of column names that comprise a private key. Typically used with "merge" write disposition to deduplicate loaded data.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        merge_key (TTableHintTemplate[TColumnNames], optional): A column name or a list of column names that define a merge key. Typically used with "merge" write disposition to remove overlapping data ranges ie. to keep a single record for a given day.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        schema_contract (TTableHintTemplate[TSchemaContract], optional): Schema contract settings that will be applied to all resources of this source (if not overridden in the resource itself)

        table_format (TTableHintTemplate[TTableFormat], optional): Defines the storage format of the table. Currently only "iceberg" is supported on Athena, and "delta" on the filesystem.
            Other destinations ignore this hint.

        file_format (TTableHintTemplate[TFileFormat], optional): Format of the file in which resource data is stored. Useful when importing external files. Use `preferred` to force
            a file format that is preferred by the destination used. This setting superseded the `load_file_format` passed to pipeline `run` method.

        references (TTableHintTemplate[TTableReferenceParam], optional): A list of references to other table's columns.
            A list in the form of `[{'referenced_table': 'other_table', 'columns': ['other_col1', 'other_col2'], 'referenced_columns': ['col1', 'col2']}]`.
            Table and column names will be normalized according to the configured naming convention.

        nested_hints (Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]], optional): Hints for nested tables created by this resource.

        selected (bool, optional): When `True` `dlt pipeline` will extract and load this resource, if `False`, the resource will be ignored.

        spec (Type[BaseConfiguration], optional): A specification of configuration and secret values required by the source.

        parallelized (bool, optional): If `True`, the resource generator will be extracted in parallel with other resources.
            Transformers that return items are also parallelized. Defaults to `False`.

        incremental (Optional[TIncrementalConfig], optional): An incremental configuration for the resource.

        _impl_cls (Type[TDltResourceImpl], optional): A custom implementation of DltResource, may be also used to providing just a typing stub

        section (Optional[TTableHintTemplate[str]], optional): Configuration section that comes right after 'sources` in default layout. If not present, the current python module name will be used.
            Default layout is `sources.<section>.<name>.<key_name>`. Note that resource section is used only when a single resource is passed to the pipeline.

        _base_spec (Type[BaseConfiguration], optional): A base spec used to which spec derived from resource function arguments is added

        standalone (bool, optional): Deprecated. Past functionality got merged into regular resource

        data_from (TUnboundDltResource, optional): Allows to pipe data from one resource to another to build multi-step pipelines.

    Raises:
        ResourceNameMissing: indicates that name of the resource cannot be inferred from the `data` being passed.
        InvalidResourceDataType: indicates that the `data` argument cannot be converted into `dlt resource`

    Returns:
        Any: TDltResourceImpl instance which may be loaded, iterated or combined with other resources into a pipeline.
    """

    def make_resource(_name: str, _section: str, _data: Any) -> TDltResourceImpl:
        table_template = make_hints(
            table_name,
            write_disposition=write_disposition or DEFAULT_WRITE_DISPOSITION,
            columns=columns,
            primary_key=primary_key,
            merge_key=merge_key,
            schema_contract=schema_contract,
            table_format=table_format,
            file_format=file_format,
            references=references,
            nested_hints=nested_hints,
            incremental=incremental,
        )
        resource = _impl_cls.from_data(
            _data,
            _name,
            _section,
            table_template,
            selected,
            cast(DltResource, data_from),
            True,
        )

        if incremental:
            # Reset the flag to allow overriding by incremental argument
            resource.incremental._from_hints = False
        # If custom nesting level was specified then
        # we need to add it to table hints so that
        # later in normalizer dlt/common/normalizers/json/relational.py
        # we can override max_nesting level for the given table
        if max_table_nesting is not None:
            resource.max_table_nesting = max_table_nesting
        if parallelized:
            return resource.parallelize()
        return resource

    def decorator(
        f: Callable[TResourceFunParams, Any],
    ) -> TDltResourceImpl:
        if not callable(f):
            if data_from:
                # raise more descriptive exception if we construct transformer
                raise InvalidTransformerDataTypeGeneratorFunctionRequired(
                    name or "<no name>", f, type(f)
                )
            raise ResourceFunctionExpected(name or "<no name>", f, type(f))

        # allow for name and section to be created at runtime via dynstr which will
        # be evaluated when resource will be parametrized/called
        func_module = inspect.getmodule(f)
        if callable(name):
            resource_name: str = dynstr(get_callable_name(f), name)
        else:
            resource_name = name or get_callable_name(f)
        if callable(section):
            source_section: str = dynstr(_get_source_section_name(func_module), section)
        else:
            source_section = section or _get_source_section_name(func_module)

        is_inner_resource = is_inner_callable(f)
        if spec is None:
            # autodetect spec
            SPEC, resolvable_fields = spec_from_signature(f, inspect.signature(f), base=_base_spec)
        else:
            SPEC, resolvable_fields = spec, spec.get_resolvable_fields()

        # add spec to f
        set_fun_spec(f, SPEC)

        if is_inner_resource and len(resolvable_fields) > 0 and data_from is not None:
            # warn for parametrized inner transformers due to injection costs
            # we do not warn on top level ones because that's normal pattern here
            logger.info(
                f"Transformer {resource_name} in section {source_section} is defined over an inner"
                " function and has additional arguments that will be injected from the"
                " configuration. Note that transformers  are called for each data item (typically"
                " batch or page) and configuration injection is costly .Use the dlt.source to get"
                " the required configuration and pass them explicitly to your source."
            )

        factory = None
        r_ = make_resource(resource_name, source_section, f)
        # register non inner resources as source with single resource in it
        if not is_inner_resource:
            # a source function for the source wrapper, args that go to source are forwarded
            # to a single resource within
            def _source(
                name_ovr: str, section_ovr: str, args: Tuple[Any, ...], kwargs: Dict[str, Any]
            ) -> DltResource:
                return r_.with_name(name_ovr, new_section=section_ovr)(*args, **kwargs)

            # make the source module same as original resource
            _source.__qualname__ = f.__qualname__
            _source.__module__ = f.__module__
            # setup our special single resource source
            factory = (
                DltSourceFactoryWrapper[Any, DltSource]()
                .clone(
                    name=resource_name,
                    section=source_section,
                    spec=_base_spec,
                    _impl_cls=_DltSingleSource,
                )
                .bind(_source)
            )
            # mod the reference to keep the right spec
            factory.ref.SPEC = SPEC

        # associate source factory with the decorated function for the standalone=True resource
        # this provides access to standalone resources in the same way as to sources via SourceReference
        r_._factory = factory  # type: ignore[attr-defined]

        return r_

    # if data is callable or none use decorator
    if data is None:
        # we're called with parens.
        return decorator

    if callable(data):
        return decorator(data)
    else:
        assert not callable(name)
        assert not callable(section)

        # take name from the generator
        source_section: str = None
        if inspect.isgenerator(data):
            name = name or get_callable_name(data)  # type: ignore
            func_module = inspect.getmodule(data.gi_frame)
            source_section = section or _get_source_section_name(func_module)
        return make_resource(name, source_section, data)


@overload
def transformer(
    f: None = ...,
    /,
    data_from: TUnboundDltResource = DltResource.Empty,
    name: TTableHintTemplate[str] = None,
    table_name: TTableHintTemplate[str] = None,
    max_table_nesting: int = None,
    write_disposition: TTableHintTemplate[TWriteDisposition] = None,
    columns: TTableHintTemplate[TAnySchemaColumns] = None,
    primary_key: TTableHintTemplate[TColumnNames] = None,
    merge_key: TTableHintTemplate[TColumnNames] = None,
    schema_contract: TTableHintTemplate[TSchemaContract] = None,
    table_format: TTableHintTemplate[TTableFormat] = None,
    file_format: TTableHintTemplate[TFileFormat] = None,
    references: TTableHintTemplate[TTableReferenceParam] = None,
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    section: Optional[TTableHintTemplate[str]] = None,
    standalone: bool = None,
) -> Callable[
    [Callable[Concatenate[TDataItem, TResourceFunParams], Any]],
    ResourceFactory[TResourceFunParams, DltResource],
]: ...


@overload
def transformer(
    f: Callable[Concatenate[TDataItem, TResourceFunParams], Any],
    /,
    data_from: TUnboundDltResource = DltResource.Empty,
    name: TTableHintTemplate[str] = None,
    table_name: TTableHintTemplate[str] = None,
    max_table_nesting: int = None,
    write_disposition: TTableHintTemplate[TWriteDisposition] = None,
    columns: TTableHintTemplate[TAnySchemaColumns] = None,
    primary_key: TTableHintTemplate[TColumnNames] = None,
    merge_key: TTableHintTemplate[TColumnNames] = None,
    schema_contract: TTableHintTemplate[TSchemaContract] = None,
    table_format: TTableHintTemplate[TTableFormat] = None,
    file_format: TTableHintTemplate[TFileFormat] = None,
    references: TTableHintTemplate[TTableReferenceParam] = None,
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    section: Optional[TTableHintTemplate[str]] = None,
    standalone: bool = None,
) -> ResourceFactory[TResourceFunParams, DltResource]: ...


def transformer(
    f: Optional[Callable[Concatenate[TDataItem, TResourceFunParams], Any]] = None,
    /,
    data_from: TUnboundDltResource = DltResource.Empty,
    name: TTableHintTemplate[str] = None,
    table_name: TTableHintTemplate[str] = None,
    max_table_nesting: int = None,
    write_disposition: TTableHintTemplate[TWriteDisposition] = None,
    columns: TTableHintTemplate[TAnySchemaColumns] = None,
    primary_key: TTableHintTemplate[TColumnNames] = None,
    merge_key: TTableHintTemplate[TColumnNames] = None,
    schema_contract: TTableHintTemplate[TSchemaContract] = None,
    table_format: TTableHintTemplate[TTableFormat] = None,
    file_format: TTableHintTemplate[TFileFormat] = None,
    references: TTableHintTemplate[TTableReferenceParam] = None,
    nested_hints: Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None,
    parallelized: bool = False,
    section: Optional[TTableHintTemplate[str]] = None,
    standalone: bool = None,
    _impl_cls: Type[TDltResourceImpl] = DltResource,  # type: ignore[assignment]
) -> Any:
    """A form of `dlt resource` that takes input from other resources via `data_from` argument in order to enrich or transform the data.

    The decorated function `f` must take at least one argument of type TDataItems (a single item or list of items depending on the resource `data_from`). `dlt` will pass
    metadata associated with the data item if argument with name `meta` is present. Otherwise, transformer function may take more arguments and be parametrized
    like the resources.

    You can bind the transformer early by specifying resource in `data_from` when the transformer is created or create dynamic bindings later with | operator
    which is demonstrated in example below:

    Example:
    >>> @dlt.resource
    >>> def players(title, chess_url=dlt.config.value):
    >>>     r = requests.get(f"{chess_url}titled/{title}")
    >>>     yield r.json()["players"]  # returns list of player names
    >>>
    >>> # this resource takes data from players and returns profiles
    >>> @dlt.transformer(write_disposition="replace")
    >>> def player_profile(player: Any) -> Iterator[TDataItems]:
    >>>     r = requests.get(f"{chess_url}player/{player}")
    >>>     r.raise_for_status()
    >>>     yield r.json()
    >>>
    >>> # pipes the data from players into player profile to produce a list of player profiles
    >>> list(players("GM") | player_profile)

    Args:
        f (Optional[Callable[Concatenate[TDataItem, TResourceFunParams], Any]]): a function taking minimum one argument of TDataItems type which will receive data yielded from `data_from` resource.

        data_from (TUnboundDltResource, optional): a resource that will send data to the decorated function `f`

        name (TTableHintTemplate[str], optional): A name of the resource that by default also becomes the name of the table to which the data is loaded.
            If not present, the name of the decorated function will be used.

        table_name (TTableHintTemplate[str], optional): An table name, if different from `name`.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        max_table_nesting (int, optional): A schema hint that sets the maximum depth of nested table above which the remaining nodes are loaded as structs or JSON.

        write_disposition (TTableHintTemplate[TWriteDisposition], optional): Controls how to write data to a table. `append` will always add new data at the end of the table. `replace` will replace existing data with new data. `skip` will prevent data from loading. "merge" will deduplicate and merge data based on "primary_key" and "merge_key" hints. Defaults to "append".
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        columns (TTableHintTemplate[TAnySchemaColumns], optional): A list, dict or pydantic model of column schemas. Typed dictionary describing column names, data types, write disposition and performance hints that gives you full control over the created table schema.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        primary_key (TTableHintTemplate[TColumnNames], optional): A column name or a list of column names that comprise a private key. Typically used with "merge" write disposition to deduplicate loaded data.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        merge_key (TTableHintTemplate[TColumnNames], optional): A column name or a list of column names that define a merge key. Typically used with "merge" write disposition to remove overlapping data ranges ie. to keep a single record for a given day.
            This argument also accepts a callable that is used to dynamically create tables for stream-like resources yielding many datatypes.

        schema_contract (TTableHintTemplate[TSchemaContract], optional): Schema contract settings that will be applied to all resources of this source (if not overridden in the resource itself)

        table_format (TTableHintTemplate[TTableFormat], optional): Defines the storage format of the table. Currently only "iceberg" is supported on Athena, and "delta" on the filesystem.
            Other destinations ignore this hint.

        file_format (TTableHintTemplate[TFileFormat], optional): Format of the file in which resource data is stored. Useful when importing external files. Use `preferred` to force
            a file format that is preferred by the destination used. This setting superseded the `load_file_format` passed to pipeline `run` method.

        references (TTableHintTemplate[TTableReferenceParam], optional): A list of references to other table's columns.
            A list in the form of `[{'referenced_table': 'other_table', 'columns': ['other_col1', 'other_col2'], 'referenced_columns': ['col1', 'col2']}]`.
            Table and column names will be normalized according to the configured naming convention.

        nested_hints (Optional[TTableHintTemplate[Dict[TTableNames, TResourceNestedHints]]], optional): Hints for nested tables created by this resource.

        selected (bool, optional): When `True` `dlt pipeline` will extract and load this resource, if `False`, the resource will be ignored.

        spec (Type[BaseConfiguration], optional): A specification of configuration and secret values required by the source.

        parallelized (bool, optional): When `True` the resource will be loaded in parallel.

        section (Optional[TTableHintTemplate[str]], optional): Configuration section that comes right after 'sources` in default layout. If not present, the current python module name will be used.
            Default layout is `sources.<section>.<name>.<key_name>`. Note that resource section is used only when a single resource is passed to the pipeline.

        standalone (bool, optional): Deprecated. Past functionality got merged into regular resource

        _impl_cls (Type[TDltResourceImpl], optional): A custom implementation of DltResource, may be also used to providing just a typing stub

    Raises:
        ResourceNameMissing: indicates that name of the resource cannot be inferred from the `data` being passed.
        InvalidResourceDataType: indicates that the `data` argument cannot be converted into `dlt resource`

    Returns:
        Any: TDltResourceImpl instance which may be loaded, iterated or combined with other resources into a pipeline.
    """
    if isinstance(f, DltResource):
        raise ValueError(
            "Please pass `data_from=` argument as keyword argument. The only positional argument to"
            " transformer is the decorated function"
        )

    return resource(  # type: ignore
        f,
        name=name,
        table_name=table_name,
        max_table_nesting=max_table_nesting,
        write_disposition=write_disposition,
        columns=columns,
        primary_key=primary_key,
        merge_key=merge_key,
        schema_contract=schema_contract,
        table_format=table_format,
        file_format=file_format,
        references=references,
        nested_hints=nested_hints,
        selected=selected,
        spec=spec,
        data_from=data_from,
        parallelized=parallelized,
        _impl_cls=_impl_cls,
        section=section,
    )


def _maybe_load_schema_for_callable(f: AnyFun, name: str) -> Optional[Schema]:
    if not inspect.isfunction(f):
        f = f.__class__
    try:
        file = inspect.getsourcefile(f)
        if file:
            schema = SchemaStorage.load_schema_file(
                os.path.dirname(file), name, remove_processing_hints=True
            )
            schema.update_normalizers()
            return schema
    except SchemaNotFoundError:
        pass
    return None


def _get_source_section_name(m: ModuleType) -> str:
    """Gets the source section name (as in SOURCES (section, name) tuple) from __source_name__ of the module `m` or from its name"""
    if m is None:
        return None
    if hasattr(m, "__source_name__"):
        return cast(str, m.__source_name__)
    return get_module_name(m)


def get_source_schema() -> Schema:
    """Should be executed from inside the function decorated with @dlt.resource

    Returns:
        Schema: The current writeable source schema
    """
    try:
        return Container()[SourceSchemaInjectableContext].schema
    except ContextDefaultCannotBeCreated:
        raise CurrentSourceSchemaNotAvailable()


def get_source() -> DltSource:
    """Should be executed from inside the function decorated with @dlt.resource

    Returns:
        DltSource: The current writable source object
    """
    try:
        return Container()[SourceInjectableContext].source
    except ContextDefaultCannotBeCreated:
        raise CurrentSourceNotAvailable()


def get_resource() -> DltResource:
    """Should be executed from inside the function decorated with @dlt.resource

    Returns:
        DltResource: The resource object to which the currently executing pipe belongs
    """
    source = get_source()
    return source.resources.with_pipe(get_current_pipe())


def get_resource_metrics() -> Dict[str, Any]:
    """Should be executed from inside the function decorated with @dlt.resource

    Returns:
        Dict[str, Any]: The customizable metrics dictionary
    """
    return get_resource().custom_metrics


TBoundItems = TypeVar("TBoundItems", bound=TDataItems)
TDeferred = Callable[[], TBoundItems]
TDeferredFunParams = ParamSpec("TDeferredFunParams")


def defer(
    f: Callable[TDeferredFunParams, TBoundItems],
) -> Callable[TDeferredFunParams, TDeferred[TBoundItems]]:
    @wraps(f)
    def _wrap(*args: Any, **kwargs: Any) -> TDeferred[TBoundItems]:
        def _curry() -> TBoundItems:
            return f(*args, **kwargs)

        return _curry

    # copy the signature from f to _wrap, but change the return type
    sig = inspect.signature(f)
    return_annotation = (
        Callable[[], sig.return_annotation]
        if sig.return_annotation is not inspect.Parameter.empty
        else None
    )
    if return_annotation:
        new_sig = sig.replace(return_annotation=return_annotation)
        setattr(_wrap, "__signature__", new_sig)

    return _wrap



from typing import cast
import pyarrow as pa
import pyarrow.parquet as pq
from lancedb import DBConnection

from dlt.common.destination.client import (
    RunnableLoadJob,
    HasFollowupJobs,
)
from dlt.common.schema.typing import (
    TWriteDisposition,
    TTableSchema,
)
from dlt.common.schema.utils import is_nested_table
from dlt.common.storages import ParsedLoadJobFileName
from dlt.destinations.impl.lancedb.schema import (
    TArrowSchema,
    TTableLineage,
    TableJob,
)
from dlt.destinations.impl.lancedb.utils import (
    get_canonical_vector_database_doc_id_merge_key,
    create_in_filter,
    write_records,
)
from dlt.destinations.job_impl import ReferenceFollowupJobRequest
from dlt.destinations.sql_jobs import SqlMergeFollowupJob


class LanceDBLoadJob(RunnableLoadJob, HasFollowupJobs):
    arrow_schema: TArrowSchema

    def __init__(
        self,
        file_path: str,
        table_schema: TTableSchema,
    ) -> None:
        from dlt.destinations.impl.lancedb.lancedb_client import LanceDBClient

        super().__init__(file_path)
        self._job_client: "LanceDBClient" = None
        self._table_schema: TTableSchema = table_schema

    def run(self) -> None:
        db_client: DBConnection = self._job_client.db_client
        fq_table_name: str = self._job_client.make_qualified_table_name(self._table_schema["name"])
        write_disposition: TWriteDisposition = cast(
            TWriteDisposition, self._load_table.get("write_disposition", "append")
        )

        merge_key: str = None
        if write_disposition == "merge":
            # use deterministic and unique id as a merge column (to perform classical upsert)
            # NOTE: upsert strategy generates deterministic row_key both for root and nested tables
            merge_key = SqlMergeFollowupJob.get_row_key_col(
                [self._load_table],
                self._load_table,
                self._job_client.dataset_name,
                self._job_client.dataset_name,
            )

        with open(self._file_path, mode="rb") as f:
            arrow_table: pa.Table = pq.read_table(f)

        write_records(
            arrow_table,
            db_client=db_client,
            vector_field_name=self._job_client.config.vector_field_name,
            table_name=fq_table_name,
            write_disposition=write_disposition,
            merge_key=merge_key,
        )


class LanceDBRemoveOrphansJob(RunnableLoadJob):
    def __init__(
        self,
        file_path: str,
    ) -> None:
        from dlt.destinations.impl.lancedb.lancedb_client import LanceDBClient

        super().__init__(file_path)
        self._job_client: "LanceDBClient" = None
        self.references = ReferenceFollowupJobRequest.resolve_references(file_path)

    def run(self) -> None:
        db_client: DBConnection = self._job_client.db_client
        table_lineage: TTableLineage = [
            TableJob(
                table_schema=self._job_client.prepare_load_table(
                    ParsedLoadJobFileName.parse(file_path_).table_name
                ),
                table_name=ParsedLoadJobFileName.parse(file_path_).table_name,
                file_path=file_path_,
            )
            for file_path_ in self.references
        ]

        for job in table_lineage:
            target_is_root_table = not is_nested_table(job.table_schema)
            fq_table_name = self._job_client.make_qualified_table_name(job.table_name)
            file_path = job.file_path
            with open(file_path, mode="rb") as f:
                payload_arrow_table: pa.Table = pq.read_table(f)

            if target_is_root_table:
                canonical_doc_id_field = get_canonical_vector_database_doc_id_merge_key(
                    job.table_schema
                )
                # delete all records with load id different than load id of payload_arrow_table
                # that have docs ids in payload_arrow_table (orphaned rows)
                delete_condition = create_in_filter(
                    canonical_doc_id_field, payload_arrow_table[canonical_doc_id_field]
                )
                # TODO: raise if dlt_load_id not present in payload_arrow_table. most probably
                #   arrow tables are used but normalizer skips _dlt_load_id
                dlt_load_id = self._schema.data_item_normalizer.c_dlt_load_id  # type: ignore[attr-defined]
                merge_key = dlt_load_id

            else:
                dlt_id = SqlMergeFollowupJob.get_row_key_col(
                    [job.table_schema],
                    job.table_schema,
                    self._job_client.dataset_name,
                    self._job_client.dataset_name,
                )
                dlt_root_id = SqlMergeFollowupJob.get_root_key_col(
                    [job.table_schema],
                    job.table_schema,
                    self._job_client.dataset_name,
                    self._job_client.dataset_name,
                )
                # delete all records with dlt id not in payload_arrow_table
                # that have root key id in payload_arrow_table (orphaned rows)
                delete_condition = create_in_filter(
                    dlt_root_id,
                    payload_arrow_table[dlt_root_id],
                )
                merge_key = dlt_id

            write_records(
                payload_arrow_table,
                db_client=db_client,
                vector_field_name=self._job_client.config.vector_field_name,
                table_name=fq_table_name,
                write_disposition="merge",
                merge_key=merge_key,
                remove_orphans=True,
                delete_condition=delete_condition,
            )



import contextlib
import logging
import traceback
from logging import LogRecord, Logger
from typing import Any, Mapping, Iterator, Protocol, Callable

LOGGER: Logger = None
_MESSAGE_LOGGING_METHODS = {
    "debug",
    "info",
    "warning",
    "warn",
    "error",
    "critical",
    "exception",
    "fatal",
}


class LogMethod(Protocol):
    def __call__(self, *args: Any, **kwds: Any) -> Any: ...


def __getattr__(name: str) -> LogMethod:
    """Forwards log method calls (debug, info, error etc.) to LOGGER"""

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if LOGGER is None:
            return None

        target: Callable[..., Any] = getattr(LOGGER, name)

        if name in _MESSAGE_LOGGING_METHODS:
            # skip stack frames when displaying log so the original logging frame is displayed
            stacklevel = 2
            if name == "exception":
                # exception has one more frame
                stacklevel = 3
            kwargs["stacklevel"] = stacklevel
        return target(*args, **kwargs)

    wrapper.__name__ = name
    return wrapper


def metrics(name: str, extra: Mapping[str, Any], stacklevel: int = 1) -> None:
    """Forwards metrics call to LOGGER"""
    if LOGGER:
        LOGGER.info(name, extra=extra, stacklevel=stacklevel)


@contextlib.contextmanager
def suppress_and_warn(msg: str) -> Iterator[None]:
    try:
        yield
    except Exception:
        LOGGER.warning(msg, exc_info=True)


def is_logging() -> bool:
    return LOGGER is not None


def log_level() -> str:
    if not LOGGER:
        raise RuntimeError("Logger not initialized")
    return logging.getLevelName(LOGGER.level)


def is_json_logging(log_format: str) -> bool:
    return log_format == "JSON"


def pretty_format_exception() -> str:
    return traceback.format_exc()


class _MetricsFormatter(logging.Formatter):
    def format(self, record: LogRecord) -> str:  # noqa: A003
        from dlt.common.json import json

        s = super(_MetricsFormatter, self).format(record)
        # dump metrics dictionary nicely
        if "metrics" in record.__dict__:
            s = s + ": " + json.dumps(record.__dict__["metrics"])
        return s


def _create_logger(
    logger_name: str, level: str, fmt: str, component: str, version: Mapping[str, str]
) -> Logger:
    if logger_name == "root":
        logging.basicConfig(level=level)
        handler = logging.getLogger().handlers[0]
        logger = logging.getLogger()
    else:
        logger = logging.getLogger(logger_name)
        logger.propagate = False
        logger.setLevel(level)
        # get or create logging handler, we log to stderr by default
        handler = next(iter(logger.handlers), logging.StreamHandler())
        logger.addHandler(handler)

    # set right formatter
    if is_json_logging(fmt):
        from dlt.common.runtime import json_logging

        class _CustomJsonFormatter(json_logging.JSONLogFormatter):
            version: Mapping[str, str] = None

            def _format_log_object(self, record: LogRecord) -> Any:
                json_log_object = super(_CustomJsonFormatter, self)._format_log_object(record)
                if self.version:
                    json_log_object.update({"version": self.version})
                return json_log_object

        json_logging.COMPONENT_NAME = component
        if "process" in json_logging.RECORD_ATTR_SKIP_LIST:
            json_logging.RECORD_ATTR_SKIP_LIST.remove("process")
        # set version as class variable as we cannot pass custom constructor parameters
        _CustomJsonFormatter.version = version
        # the only thing method above effectively does is to replace the formatter
        json_logging.init(custom_formatter=_CustomJsonFormatter)
        if logger_name == "root":
            json_logging.config_root_logger()
    else:
        handler.setFormatter(_MetricsFormatter(fmt=fmt, style="{"))

    return logger


def _delete_current_logger() -> None:
    if not LOGGER:
        return

    for handler in LOGGER.handlers[:]:
        LOGGER.removeHandler(handler)

    LOGGER.disabled = True
    LOGGER.propagate = False



import os
from typing import Optional, Any, Type
import yaml
from enum import Enum
from importlib.metadata import version as pkg_version

from dlt.version import DLT_PKG_NAME
from dlt.common.configuration.providers import SECRETS_TOML, SECRETS_TOML_KEY
from dlt.common.configuration.utils import serialize_value
from dlt.common.destination.reference import Destination
from dlt.common.libs.git import is_dirty

from dlt._workspace.cli import utils
from dlt._workspace.cli import echo as fmt
from dlt._workspace.cli._deploy_command_helpers import (
    BaseDeployment,
    ask_files_overwrite,
    generate_pip_freeze,
    github_origin_to_url,
    serialize_templated_yaml,
    wrap_template_str,
    get_schedule_description,
)
from dlt._workspace.cli.exceptions import CliCommandException, PipelineWasNotRun
from dlt._workspace.cli._urls import DLT_DEPLOY_DOCS_URL

from dlt.pipeline.exceptions import CannotRestorePipelineException

REQUIREMENTS_GITHUB_ACTION = "requirements_github_action.txt"
DLT_AIRFLOW_GCP_DOCS_URL = (
    "https://dlthub.com/docs/walkthroughs/deploy-a-pipeline/deploy-with-airflow-composer"
)
AIRFLOW_GETTING_STARTED = "https://airflow.apache.org/docs/apache-airflow/stable/start.html"
AIRFLOW_DAG_TEMPLATE_SCRIPT = "dag_template.py"
AIRFLOW_CLOUDBUILD_YAML = "cloudbuild.yaml"
COMMAND_REPO_LOCATION = "https://github.com/dlt-hub/dlt-%s-template.git"
COMMAND_DEPLOY_REPO_LOCATION = COMMAND_REPO_LOCATION % "deploy"


class DeploymentMethods(Enum):
    github_actions = "github-action"
    airflow_composer = "airflow-composer"


class SecretFormats(Enum):
    env = "env"
    toml = "toml"


def deploy_command(
    pipeline_script_path: str,
    deployment_method: str,
    repo_location: str,
    branch: Optional[str] = None,
    **kwargs: Any,
) -> None:
    # get current repo local folder
    deployment_class: Type[BaseDeployment] = None
    if deployment_method == DeploymentMethods.github_actions.value:
        deployment_class = GithubActionDeployment
    elif deployment_method == DeploymentMethods.airflow_composer.value:
        deployment_class = AirflowDeployment
    else:
        raise ValueError(
            f"Deployment method '{deployment_method}' is not supported. Only"
            f" {', '.join([m.value for m in DeploymentMethods])} are available.'"
        )
    # command no longer needed
    kwargs.pop("command", None)
    deployment_class(
        pipeline_script_path=pipeline_script_path, location=repo_location, branch=branch, **kwargs
    ).run_deployment()


class GithubActionDeployment(BaseDeployment):
    def __init__(
        self,
        pipeline_script_path: str,
        location: str,
        schedule: Optional[str],
        run_on_push: bool = False,
        run_manually: bool = False,
        branch: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(pipeline_script_path, location, branch)
        self.schedule = schedule
        self.run_on_push = run_on_push
        self.run_manually = run_manually
        self.schedule_description: Optional[str]

    def _generate_workflow(self, *args: Optional[Any]) -> None:
        self.deployment_method = DeploymentMethods.github_actions.value
        # validate schedule
        self.schedule_description = get_schedule_description(self.schedule)
        if self.schedule_description is None:
            # TODO: move that check to _dlt and some intelligent help message on missing arg
            raise ValueError(
                f"Setting 'schedule' for '{self.deployment_method}' is required! Use deploy command"
                f" as 'dlt deploy chess.py {self.deployment_method} --schedule \"*/30 * * * *\"'."
            )
        workflow = self._create_new_workflow()
        serialized_workflow = serialize_templated_yaml(workflow)
        serialized_workflow_name = f"run_{self.state['pipeline_name']}_workflow.yml"
        self.artifacts["serialized_workflow"] = serialized_workflow
        self.artifacts["serialized_workflow_name"] = serialized_workflow_name

        # pip freeze special requirements file
        with self.template_storage.open_file(
            os.path.join(self.deployment_method, "requirements_blacklist.txt")
        ) as f:
            requirements_blacklist = f.readlines()
        requirements_txt = generate_pip_freeze(requirements_blacklist, REQUIREMENTS_GITHUB_ACTION)
        requirements_txt_name = REQUIREMENTS_GITHUB_ACTION
        # if repo_storage.has_file(utils.REQUIREMENTS_TXT):
        self.artifacts["requirements_txt"] = requirements_txt
        self.artifacts["requirements_txt_name"] = requirements_txt_name

    def _make_modification(self) -> None:
        if not self.repo_storage.has_folder(utils.GITHUB_WORKFLOWS_DIR):
            self.repo_storage.create_folder(utils.GITHUB_WORKFLOWS_DIR)

        self.repo_storage.save(
            os.path.join(utils.GITHUB_WORKFLOWS_DIR, self.artifacts["serialized_workflow_name"]),
            self.artifacts["serialized_workflow"],
        )
        self.repo_storage.save(
            self.artifacts["requirements_txt_name"], self.artifacts["requirements_txt"]
        )

    def _create_new_workflow(self) -> Any:
        with self.template_storage.open_file(
            os.path.join(self.deployment_method, "run_pipeline_workflow.yml")
        ) as f:
            workflow = yaml.safe_load(f)
        # customize the workflow
        workflow["name"] = (
            f"Run {self.state['pipeline_name']} pipeline from {self.pipeline_script_path}"
        )
        if self.run_on_push is False:
            del workflow["on"]["push"]
        if self.run_manually is False:
            del workflow["on"]["workflow_dispatch"]
        workflow["on"]["schedule"] = [{"cron": self.schedule}]
        workflow["env"] = {}
        for env_var in self.envs:
            env_key = self.env_prov.get_key_name(env_var.key, *env_var.sections)
            # print(serialize_value(env_var.value))
            workflow["env"][env_key] = str(serialize_value(env_var.value))
        for secret_var in self.secret_envs:
            env_key = self.env_prov.get_key_name(secret_var.key, *secret_var.sections)
            workflow["env"][env_key] = wrap_template_str("secrets.%s") % env_key

        # run the correct script at the end
        last_workflow_step = workflow["jobs"]["run_pipeline"]["steps"][-1]
        assert last_workflow_step["run"] == "python pipeline.py"
        # must run in the directory of the script
        wf_run_path, wf_run_name = os.path.split(self.repo_pipeline_script_path)
        if wf_run_path:
            run_cd_cmd = f"cd '{wf_run_path}' && "
        else:
            run_cd_cmd = ""
        last_workflow_step["run"] = f"{run_cd_cmd}python '{wf_run_name}'"

        return workflow

    def _echo_instructions(self, *args: Optional[Any]) -> None:
        fmt.echo(
            "Your %s deployment for pipeline %s in script %s is ready!"
            % (
                fmt.bold(self.deployment_method),
                fmt.bold(self.state["pipeline_name"]),
                fmt.bold(self.pipeline_script_path),
            )
        )
        #  It contains all relevant configurations and references to credentials that are needed to run the pipeline
        fmt.echo(
            "* A github workflow file %s was created in %s."
            % (
                fmt.bold(self.artifacts["serialized_workflow_name"]),
                fmt.bold(utils.GITHUB_WORKFLOWS_DIR),
            )
        )
        fmt.echo(
            "* The schedule with which the pipeline is run is: %s.%s%s"
            % (
                fmt.bold(self.schedule_description),
                " You can also run the pipeline manually." if self.run_manually else "",
                (
                    " Pipeline will also run on each push to the repository."
                    if self.run_on_push
                    else ""
                ),
            )
        )
        fmt.echo(
            "* The dependencies that will be used to run the pipeline are stored in %s. If you"
            " change add more dependencies, remember to refresh your deployment by running the same"
            " 'deploy' command again."
            % fmt.bold(self.artifacts["requirements_txt_name"])
        )
        fmt.echo()
        if len(self.secret_envs) == 0 and len(self.envs) == 0:
            fmt.echo("1. Your pipeline does not seem to need any secrets.")
        else:
            fmt.echo(
                "You should now add the secrets to github repository secrets, commit and push the"
                " pipeline files to github."
            )
            fmt.echo(
                "1. Add the following secret values (typically stored in %s): \n%s\nin %s"
                % (
                    fmt.bold(utils.make_dlt_settings_path(SECRETS_TOML)),
                    fmt.bold(
                        "\n".join(
                            self.env_prov.get_key_name(s_v.key, *s_v.sections)
                            for s_v in self.secret_envs
                        )
                    ),
                    fmt.bold(github_origin_to_url(self.origin, "/settings/secrets/actions")),
                )
            )
            fmt.echo()
            self._echo_secrets()

        fmt.echo(
            "2. Add stage deployment files to commit. Use your Git UI or the following command"
        )
        new_req_path = self.repo_storage.from_relative_path_to_wd(
            self.artifacts["requirements_txt_name"]
        )
        new_workflow_path = self.repo_storage.from_relative_path_to_wd(
            os.path.join(utils.GITHUB_WORKFLOWS_DIR, self.artifacts["serialized_workflow_name"])
        )
        fmt.echo(fmt.bold(f"git add {new_req_path} {new_workflow_path}"))
        fmt.echo()
        fmt.echo("3. Commit the files above. Use your Git UI or the following command")
        fmt.echo(
            fmt.bold(
                f"git commit -m 'run {self.state['pipeline_name']} pipeline with github action'"
            )
        )
        if is_dirty(self.repo):
            fmt.warning(
                "You have modified files in your repository. Do not forget to push changes to your"
                " pipeline script as well!"
            )
        fmt.echo()
        fmt.echo("4. Push changes to github. Use your Git UI or the following command")
        fmt.echo(fmt.bold("git push origin"))
        fmt.echo()
        fmt.echo("5. Your pipeline should be running! You can monitor it here:")
        fmt.echo(
            fmt.bold(
                github_origin_to_url(
                    self.origin, f"/actions/workflows/{self.artifacts['serialized_workflow_name']}"
                )
            )
        )


class AirflowDeployment(BaseDeployment):
    def __init__(
        self,
        pipeline_script_path: str,
        location: str,
        branch: Optional[str] = None,
        secrets_format: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(pipeline_script_path, location, branch)
        self.secrets_format = secrets_format

    def _generate_workflow(self, *args: Optional[Any]) -> None:
        self.deployment_method = DeploymentMethods.airflow_composer.value

        req_dep = f"{DLT_PKG_NAME}[{Destination.to_name(self.state['destination_type'])}]"
        req_dep_line = f"{req_dep}>={pkg_version(DLT_PKG_NAME)}"

        self.artifacts["requirements_txt"] = req_dep_line

        dag_script_name = f"dag_{self.state['pipeline_name']}.py"
        self.artifacts["dag_script_name"] = dag_script_name

        cloudbuild_file = self.template_storage.load(
            os.path.join(self.deployment_method, AIRFLOW_CLOUDBUILD_YAML)
        )
        self.artifacts["cloudbuild_file"] = cloudbuild_file

        # TODO: rewrite dag file to at least set the schedule
        dag_file = self.template_storage.load(
            os.path.join(self.deployment_method, AIRFLOW_DAG_TEMPLATE_SCRIPT)
        )
        self.artifacts["dag_file"] = dag_file

        # ask user if to overwrite the files
        dest_dag_script = os.path.join(utils.AIRFLOW_DAGS_FOLDER, dag_script_name)
        ask_files_overwrite([dest_dag_script])

    def _make_modification(self) -> None:
        if not self.repo_storage.has_folder(utils.AIRFLOW_DAGS_FOLDER):
            self.repo_storage.create_folder(utils.AIRFLOW_DAGS_FOLDER)

        if not self.repo_storage.has_folder(utils.AIRFLOW_BUILD_FOLDER):
            self.repo_storage.create_folder(utils.AIRFLOW_BUILD_FOLDER)

        # save cloudbuild.yaml only if not exist to allow to run the deploy command for many different pipelines
        dest_cloud_build = os.path.join(utils.AIRFLOW_BUILD_FOLDER, AIRFLOW_CLOUDBUILD_YAML)
        if not self.repo_storage.has_file(dest_cloud_build):
            self.repo_storage.save(dest_cloud_build, self.artifacts["cloudbuild_file"])
        else:
            fmt.warning(
                f"{AIRFLOW_CLOUDBUILD_YAML} already created. Delete the file and run the deploy"
                " command again to re-create."
            )

        dest_dag_script = os.path.join(utils.AIRFLOW_DAGS_FOLDER, self.artifacts["dag_script_name"])
        self.repo_storage.save(dest_dag_script, self.artifacts["dag_file"])

    def _echo_instructions(self, *args: Optional[Any]) -> None:
        fmt.echo(
            "Your %s deployment for pipeline %s is ready!"
            % (
                fmt.bold(self.deployment_method),
                fmt.bold(self.state["pipeline_name"]),
            )
        )
        fmt.echo(
            "* The airflow %s file was created in %s."
            % (fmt.bold(AIRFLOW_CLOUDBUILD_YAML), fmt.bold(utils.AIRFLOW_BUILD_FOLDER))
        )
        fmt.echo(
            "* The %s script was created in %s."
            % (fmt.bold(self.artifacts["dag_script_name"]), fmt.bold(utils.AIRFLOW_DAGS_FOLDER))
        )
        fmt.echo()

        fmt.echo("You must prepare your DAG first:")

        fmt.echo(
            "1. Import your sources in %s, configure the DAG ans tasks as needed."
            % (fmt.bold(self.artifacts["dag_script_name"]))
        )
        fmt.echo(
            "2. Test the DAG with Airflow locally .\nSee Airflow getting started: %s"
            % (fmt.bold(AIRFLOW_GETTING_STARTED))
        )
        fmt.echo()

        fmt.echo(
            "If you are planning run the pipeline with Google Cloud Composer, follow the next"
            " instructions:\n"
        )
        fmt.echo(
            "1. Read this doc and set up the Environment: %s" % (fmt.bold(DLT_AIRFLOW_GCP_DOCS_URL))
        )
        fmt.echo(
            "2. Set _BUCKET_NAME up in %s/%s file. "
            % (
                fmt.bold(utils.AIRFLOW_BUILD_FOLDER),
                fmt.bold(AIRFLOW_CLOUDBUILD_YAML),
            )
        )
        if len(self.secret_envs) == 0 and len(self.envs) == 0:
            fmt.echo("3. Your pipeline does not seem to need any secrets.")
        else:
            if self.secrets_format == SecretFormats.env.value:
                fmt.echo(
                    "3. Add the following secret values (typically stored in %s): \n%s\n%s\nin"
                    " ENVIRONMENT VARIABLES using Google Composer UI"
                    % (
                        fmt.bold(utils.make_dlt_settings_path(SECRETS_TOML)),
                        fmt.bold(
                            "\n".join(
                                self.env_prov.get_key_name(s_v.key, *s_v.sections)
                                for s_v in self.secret_envs
                            )
                        ),
                        fmt.bold(
                            "\n".join(
                                self.env_prov.get_key_name(v.key, *v.sections) for v in self.envs
                            )
                        ),
                    )
                )
                fmt.echo()
                # if fmt.confirm("Do you want to list the environment variables in the format suitable for Airflow?", default=True):
                self._echo_secrets()
                self._echo_envs()
            elif self.secrets_format == SecretFormats.toml.value:
                # build toml
                fmt.echo(
                    "3. Add the following toml-string in the Google Composer UI as the"
                    f" {SECRETS_TOML_KEY} variable."
                )
                fmt.echo()
                self._echo_secrets_toml()
            else:
                raise ValueError(self.secrets_format)

        fmt.echo("4. Add dlt package below using Google Composer UI.")
        fmt.echo(fmt.bold(self.artifacts["requirements_txt"]))
        fmt.note(
            "You may need to add more packages ie. when your source requires additional"
            " dependencies"
        )
        fmt.echo("5. Commit and push the pipeline files to github:")
        fmt.echo(
            "a. Add stage deployment files to commit. Use your Git UI or the following command"
        )

        dag_script_path = self.repo_storage.from_relative_path_to_wd(
            os.path.join(utils.AIRFLOW_DAGS_FOLDER, self.artifacts["dag_script_name"])
        )
        cloudbuild_path = self.repo_storage.from_relative_path_to_wd(
            os.path.join(utils.AIRFLOW_BUILD_FOLDER, AIRFLOW_CLOUDBUILD_YAML)
        )
        fmt.echo(fmt.bold(f"git add {dag_script_path} {cloudbuild_path}"))

        fmt.echo("b. Commit the files above. Use your Git UI or the following command")
        fmt.echo(
            fmt.bold(
                f"git commit -m 'initiate {self.state['pipeline_name']} pipeline with Airflow'"
            )
        )
        if is_dirty(self.repo):
            fmt.warning(
                "You have modified files in your repository. Do not forget to push changes to your"
                " pipeline script as well!"
            )
        fmt.echo("c. Push changes to github. Use your Git UI or the following command")
        fmt.echo(fmt.bold("git push origin"))
        fmt.echo("6. You should see your pipeline in Airflow.")


@utils.track_command("deploy", False, "deployment_method")
def deploy_command_wrapper(
    pipeline_script_path: str,
    deployment_method: str,
    repo_location: str,
    branch: Optional[str] = None,
    **kwargs: Any,
) -> None:
    try:
        utils.ensure_git_command("deploy")
    except Exception as ex:
        fmt.secho(str(ex), err=True, fg="red")
        raise CliCommandException(error_code=-2)
    from git import InvalidGitRepositoryError, NoSuchPathError

    try:
        deploy_command(
            pipeline_script_path=pipeline_script_path,
            deployment_method=deployment_method,
            repo_location=repo_location,
            branch=branch,
            **kwargs,
        )
    except (CannotRestorePipelineException, PipelineWasNotRun) as ex:
        fmt.note(
            "You must run the pipeline locally successfully at least once in order to deploy it."
        )
        raise CliCommandException(error_code=-3, raiseable_exception=ex)
    except InvalidGitRepositoryError:
        fmt.secho(
            "No git repository found for pipeline script %s." % fmt.bold(pipeline_script_path),
            err=True,
            fg="red",
        )
        fmt.note("If you do not have a repository yet, you can do either of:")
        fmt.note(
            "- Run the following command to initialize new repository: %s" % fmt.bold("git init")
        )
        fmt.note(
            "- Add your local code to Github as described here: %s"
            % fmt.bold(
                "https://docs.github.com/en/get-started/importing-your-projects-to-github/importing-source-code-to-github/adding-locally-hosted-code-to-github"
            )
        )
        fmt.note("Please refer to %s for further assistance" % fmt.bold(DLT_DEPLOY_DOCS_URL))
        raise CliCommandException(error_code=-4)
    except NoSuchPathError as path_ex:
        fmt.secho("The pipeline script does not exist\n%s" % str(path_ex), err=True, fg="red")
        raise CliCommandException(error_code=-5)



import os
import warnings
from typing import Any, ClassVar, Dict, List, Optional, Protocol, Sequence, Set
import pluggy
import argparse
import importlib.metadata

from dlt.common.configuration.specs.base_configuration import ContainerInjectableContext
from dlt.common.configuration.specs.pluggable_run_context import RunContextBase
from dlt.common.known_env import DLT_DISABLE_PLUGINS

hookspec = pluggy.HookspecMarker("dlt")
hookimpl = pluggy.HookimplMarker("dlt")


class PluginContext(ContainerInjectableContext):
    global_affinity: ClassVar[bool] = True

    manager: pluggy.PluginManager
    plugin_modules: List[str]

    def __init__(self) -> None:
        super().__init__()
        self.manager = pluggy.PluginManager("dlt")
        self.plugin_modules = []

        # take hookspecs from self
        from dlt.common.configuration import plugins

        self.manager.add_hookspecs(plugins)

        # NOTE: internal implementations (__plugins__.py) are declared as entrypoints in pyproject.toml
        self.plugin_modules = load_setuptools_entrypoints(self.manager)


def manager() -> pluggy.PluginManager:
    """Returns current plugin context"""
    from .container import Container

    return Container()[PluginContext].manager


def load_setuptools_entrypoints(m: pluggy.PluginManager) -> List[str]:
    """Scans setuptools distributions that are path or have name starting with `dlt`
    loads entry points in group `dlt` and instantiates them to initialize plugins.

    returns a list of names of top level modules/packages from detected entry points.
    """

    plugin_modules = []

    if os.environ.get(DLT_DISABLE_PLUGINS, "False").lower() == "false":
        distributions = list(importlib.metadata.distributions())
    else:
        # always plug itself
        distributions = [importlib.metadata.distribution("dlt")]

    for dist in distributions:
        # skip named dists that do not start with dlt-
        package_name = dist.metadata.get("Name")

        if not package_name or not package_name.startswith("dlt"):
            continue

        for ep in dist.entry_points:
            if (
                ep.group != "dlt"
                # already registered
                or m.get_plugin(ep.name)
                or m.is_blocked(ep.name)
            ):
                continue
            try:
                plugin = ep.load()
            except Exception as e:
                warnings.warn(
                    f"Plugin {ep.name} from {package_name} failed to load: {e}",
                    stacklevel=1,
                )
                continue
            m.register(plugin, name=ep.name)
            m._plugin_distinfo.append((plugin, pluggy._manager.DistFacade(dist)))
            top_module = ep.module.split(".")[0]
            if top_module not in plugin_modules:
                plugin_modules.append(top_module)

    return plugin_modules


@hookspec(firstresult=True)
def plug_run_context(
    run_dir: Optional[str], runtime_kwargs: Optional[Dict[str, Any]]
) -> Optional[RunContextBase]:
    """Spec for plugin hook that returns current run context.

    Args:
        run_dir (str): An initial run directory of the context
        runtime_kwargs: Any additional arguments passed to the context via PluggableRunContext.reload

    Returns:
        SupportsRunContext: A run context implementing SupportsRunContext protocol
    """


class SupportsCliCommand(Protocol):
    """Protocol for defining one dlt cli command"""

    command: str
    """name of the command"""
    help_string: str
    """the help string for argparse"""
    description: Optional[str]
    """the more detailed description for argparse, may inlcude markdown for the docs"""
    docs_url: Optional[str]
    """the default docs url to be printed in case of an exception"""

    def configure_parser(self, parser: argparse.ArgumentParser) -> None:
        """Configures the parser for the given argument"""
        ...

    def execute(self, args: argparse.Namespace) -> None:
        """Executes the command with the given arguments"""
        ...


@hookspec()
def plug_cli() -> SupportsCliCommand:
    """Spec for plugin hook that returns current run context."""


class SupportsMcpFeatures(Protocol):
    """Protocol for contributing MCP tools, prompts, and providers via plug_mcp hook"""

    name: str
    """unique name identifying this feature set"""
    tools: Sequence[Any]
    """tool functions or Tool objects to register"""
    prompts: Sequence[Any]
    """prompt functions or Prompt objects to register"""
    providers: Sequence[Any]
    """provider instances (e.g. SkillProvider) to register"""


@hookspec()
def plug_mcp(features: Set[str]) -> Optional[SupportsMcpFeatures]:
    """Spec for plugin hook that contributes MCP tools, prompts, and providers.

    MCP server will broadcast `features` to all registered plugins that may
    decide to return a MCP feature (combination of tools, skills and prompts)
    or not. The server collects all non-None results and registers everything on the
    FastMCP instance.

    Args:
        features: set of feature keywords the server requests
    """
